"""SC Controller - Universal HID driver. For all three universal HID devices.

Borrows bit of code and configuration from evdevdrv.
"""

from __future__ import annotations

import ctypes
import json
import logging
import os
import sys
from enum import IntEnum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from usb1 import USBDeviceHandle

	from scc.device_monitor import DeviceMonitor
	from scc.sccdaemon import SCCDaemon

from scc.constants import STICK_PAD_MAX, STICK_PAD_MIN, ControllerFlags, SCButtons
from scc.controller import Controller
from scc.drivers.evdevdrv import FIRST_BUTTON, TRIGGERS, parse_axis
from scc.drivers.usb import (
	USBDevice,
	register_hotplug_device,
	unregister_hotplug_device,
)
from scc.lib.hidparse import (
	AXES,
	GenericDesktopPage,
	GlobalItem,
	ItemType,
	LocalItem,
	MainItem,
	UsagePage,
	parse_report_descriptor,
)
from scc.paths import get_config_path
from scc.tools import find_library

log = logging.getLogger("HID")

DEV_CLASS_HID = 3
TRANSFER_TYPE_INTERRUPT = 3
LIBUSB_DT_REPORT = 0x22
AXIS_COUNT = 20  # Must match number of axis fields in HIDControllerInput and values in AxisType
BUTTON_COUNT = 32  # Must match (or be less than) number of bits in HIDControllerInput.buttons
ALLOWED_SIZES = [1, 2, 4, 8, 16, 32]
SYS_DEVICES = "/sys/devices"


BLACKLIST = [
	# List of devices known to pretend to be HID compatible but breaking horribly with HID
	# vendor, product
	(0x045E, 0x0719),  # Xbox controller
	(0x045E, 0x028E),  # Xbox wireless adapter
	(0x0738, 0x4716),  # Mad Catz, Inc controller
]

def button_to_bit(sc) -> int:
	sc, bit = int(sc), 0
	while sc and (sc & 1 == 0):
		sc >>= 1
		bit += 1
	if sc & 1 == 1:
		return bit
	return BUTTON_COUNT - 1

class HIDDrvError(Exception):
	pass


class NotHIDDevice(HIDDrvError):
	pass


class UnparsableDescriptor(HIDDrvError):
	pass


class HIDControllerInput(ctypes.Structure):
	_fields_ = [
		("buttons", ctypes.c_uint32),
		# Note: Axis order is same as in AxisType enum
		("lpad_x", ctypes.c_int32),
		("lpad_y", ctypes.c_int32),
		("rpad_x", ctypes.c_int32),
		("rpad_y", ctypes.c_int32),
		("stick_x", ctypes.c_int32),
		("stick_y", ctypes.c_int32),
		("ltrig", ctypes.c_int32),
		("rtrig", ctypes.c_int32),
		("accel_x", ctypes.c_int32),
		("accel_y", ctypes.c_int32),
		("accel_z", ctypes.c_int32),
		("gpitch", ctypes.c_int32),
		("groll", ctypes.c_int32),
		("gyaw", ctypes.c_int32),
		("q1", ctypes.c_int32),
		("q2", ctypes.c_int32),
		("q3", ctypes.c_int32),
		("q4", ctypes.c_int32),
		("cpad_x", ctypes.c_int32),
		("cpad_y", ctypes.c_int32),
	]


class AxisType(IntEnum):
	AXIS_LPAD_X = 0
	AXIS_LPAD_Y = 1
	AXIS_RPAD_X = 2
	AXIS_RPAD_Y = 3
	AXIS_STICK_X = 4
	AXIS_STICK_Y = 5
	AXIS_LTRIG = 6
	AXIS_RTRIG = 7
	AXIS_ACCEL_X = 8
	AXIS_ACCEL_Y = 9
	AXIS_ACCEL_Z = 10
	AXIS_GPITCH = 11
	AXIS_GROLL = 12
	AXIS_GYAW = 13
	AXIS_Q1 = 14
	AXIS_Q2 = 15
	AXIS_Q3 = 16
	AXIS_Q4 = 17
	AXIS_CPAD_X = 18
	AXIS_CPAD_Y = 19


class AxisMode(IntEnum):
	DISABLED = 0
	AXIS = 1
	AXIS_NO_SCALE = 2
	DPAD = 3
	HATSWITCH = 4
	DS4ACCEL = 5  # 16bit, signed, no additional math needed
	DS4GYRO = 6  # 16bit, signed, inverted
	DS4TOUCHPAD = 7  # 12bit


class AxisModeData(ctypes.Structure):
	_fields_ = [
		("button", ctypes.c_uint32),
		("scale", ctypes.c_float),
		("offset", ctypes.c_float),
		("clamp_min", ctypes.c_int),
		("clamp_max", ctypes.c_int),
		("deadzone", ctypes.c_float),
	]


class DPadModeData(ctypes.Structure):
	_fields_ = [
		("button", ctypes.c_uint32),
		("button1", ctypes.c_uint8),
		("button2", ctypes.c_uint8),
		("min", ctypes.c_int),
		("max", ctypes.c_int),
	]


class HatswitchModeData(ctypes.Structure):
	_fields_ = [
		("button", ctypes.c_uint32),
		("min", ctypes.c_int),
		("max", ctypes.c_int),
	]


class AxisDataUnion(ctypes.Union):
	_fields_ = [
		("axis", AxisModeData),
		("dpad", DPadModeData),
		("hatswitch", HatswitchModeData),
	]


class AxisData(ctypes.Structure):
	_fields_ = [
		("mode", ctypes.c_int),
		("byte_offset", ctypes.c_size_t),
		("bit_offset", ctypes.c_uint8),
		("size", ctypes.c_uint8),  # TODO: Currently unused
		("data", AxisDataUnion),
	]


class ButtonData(ctypes.Structure):
	_fields_ = [
		("enabled", ctypes.c_bool),
		("byte_offset", ctypes.c_size_t),
		("bit_offset", ctypes.c_uint8),
		("size", ctypes.c_uint8),
		("button_count", ctypes.c_uint8),
		("button_map", ctypes.c_uint8 * BUTTON_COUNT),
	]


class HIDDecoder(ctypes.Structure):
	_fields_ = [
		("axes", AxisData * AXIS_COUNT),
		("buttons", ButtonData),
		("packet_size", ctypes.c_size_t),
		("old_state", HIDControllerInput),
		("state", HIDControllerInput),
	]


HIDDecoderPtr = ctypes.POINTER(HIDDecoder)


_lib = find_library("libhiddrv")
_lib.decode.restype = bool
_lib.decode.argtypes = [HIDDecoderPtr, ctypes.c_char_p]


class HIDController(USBDevice, Controller):
	flags = (
		ControllerFlags.HAS_RSTICK
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.NO_GRIPS
	)

	def __init__(
		self,
		device: USBDevice,
		daemon: SCCDaemon,
		handle: USBDeviceHandle,
		config_file: str,
		config: dict,
		test_mode: bool = False,
	):
		USBDevice.__init__(self, device, handle)
		self._ready = False
		self.daemon = daemon
		self.config_file = config_file

		id = None
		max_size = 64
		for inter in self.device[0]:
			for setting in inter:
				if setting.getClass() == DEV_CLASS_HID:
					for endpoint in setting:
						if endpoint.getAttributes() == TRANSFER_TYPE_INTERRUPT:
							if id is None or endpoint.getAddress() > id:
								id = endpoint.getAddress()
								max_size = endpoint.getMaxPacketSize()

		if id is None:
			raise NotHIDDevice

		log.debug("Endpoint: %s", id)

		vid, pid = self.device.getVendorID(), self.device.getProductID()
		if (vid, pid) in BLACKLIST:
			raise NotHIDDevice("Blacklisted device: %x:%x", vid, pid)
		self._packet_size = 64
		self._load_hid_descriptor(config, max_size, vid, pid, test_mode)
		self.claim_by(klass=DEV_CLASS_HID, subclass=0, protocol=0)
		Controller.__init__(self)

		if test_mode:
			self.set_input_interrupt(id, self._packet_size, self.test_input)

			print("Buttons:", " ".join([str(x + FIRST_BUTTON) for x in range(self._decoder.buttons.button_count)]))
			print(
				"Axes:",
				" ".join([str(x) for x in range(len([a for a in self._decoder.axes if a.mode != AxisMode.DISABLED]))]),
			)
		else:
			self._id = self._generate_id()
			self.set_input_interrupt(id, self._packet_size, self.input)
			self.daemon.add_controller(self)
			self._ready = True

	def _load_hid_descriptor(self, config: dict, max_size: int, vid: int, pid: int, test_mode) -> None:
		hid_descriptor = HIDController.find_sys_devices_descriptor(vid, pid)
		if hid_descriptor is None:
			hid_descriptor = self.handle.getRawDescriptor(LIBUSB_DT_REPORT, 0, 512)
		open("report", "wb").write(bytes(list(hid_descriptor)))
		self._build_hid_decoder(hid_descriptor, config, max_size)
		self._packet_size = self._decoder.packet_size

	def _build_button_map(self, config: dict):
		"""Return button map readed from configuration, in format situable for HIDDecoder.buttons.button_map field.

		Generates default if config is not available.
		"""
		if config:
			# Last possible value is default "maps-to-nothing" mapping
			buttons = [BUTTON_COUNT - 1] * BUTTON_COUNT
			for keycode, value in config.get("buttons", {}).items():
				keycode = int(keycode) - FIRST_BUTTON
				if keycode < 0 or keycode >= BUTTON_COUNT:
					# Out of range
					continue
				if value in TRIGGERS:
					# Not used here
					pass
				else:
					buttons[keycode] = button_to_bit(getattr(SCButtons, value))
		else:
			buttons = list(range(BUTTON_COUNT))

		return (ctypes.c_uint8 * BUTTON_COUNT)(*buttons)

	def _build_axis_maping(self, axis, config: dict, mode=AxisMode.AXIS):
		"""Convert configuration mapping for _one_ axis to value situable for self._decoder.axes field."""
		axis_config = config.get("axes", {}).get(str(int(axis)))
		if axis_config:
			try:
				target = [x for (x, y) in HIDControllerInput._fields_].index(axis_config.get("axis")) - 1
			except Exception:
				# Maps to unknown axis
				return None, None
			cdata = parse_axis(axis_config)
			button = 0
			if AxisType(target) in (AxisType.AXIS_LPAD_X, AxisType.AXIS_LPAD_Y):
				button = SCButtons.LPADTOUCH | SCButtons.LPAD
			elif AxisType(target) in (AxisType.AXIS_RPAD_X, AxisType.AXIS_RPAD_Y):
				button = SCButtons.RPAD
			if mode == AxisMode.AXIS:
				axis_data = AxisData(
					mode=AxisMode.AXIS,
					data=AxisDataUnion(
						axis=AxisModeData(button=button, **{field: getattr(cdata, field) for field in cdata._fields}),
					),
				)
			elif mode == AxisMode.HATSWITCH:
				axis_data = AxisData(
					mode=AxisMode.HATSWITCH,
					data=AxisDataUnion(
						hatswitch=HatswitchModeData(button=button, max=axis_config["max"], min=axis_config["min"]),
					),
				)
			else:
				axis_data = AxisData(mode=AxisMode.DISABLED)
			return target, axis_data
		return None, None

	def _build_hid_decoder(self, data, config: dict, max_size: int) -> None:
		size, count, total, kind = 1, 0, 0, None
		next_axis = AxisType.AXIS_LPAD_X
		self._decoder = HIDDecoder()
		for x in parse_report_descriptor(data, True):
			if x[0] == GlobalItem.ReportSize:
				size = x[1]
			elif x[0] == GlobalItem.ReportCount:
				count = x[1]
			elif x[0] == LocalItem.Usage:
				kind = x[1]
			elif x[0] == MainItem.Input:
				if x[1] == ItemType.Constant:
					total += count * size
					log.debug("Found %s bits of nothing", count * size)
				elif x[1] == ItemType.Data:
					if kind in AXES:
						if size not in ALLOWED_SIZES:
							raise UnparsableDescriptor("Axis with invalid size (%s bits)" % (size,))
						for i in range(count):
							if next_axis < AXIS_COUNT:
								log.debug("Found axis #%s at bit %s", int(next_axis), total)
								if config:
									target, axis_data = self._build_axis_maping(next_axis, config)
									if axis_data:
										axis_data.byte_offset = total // 8
										axis_data.bit_offset = total % 8
										axis_data.size = size
										self._decoder.axes[target] = axis_data
								else:
									self._decoder.axes[next_axis] = AxisData(mode=AxisMode.AXIS_NO_SCALE)
									self._decoder.axes[next_axis].byte_offset = total // 8
									self._decoder.axes[next_axis].bit_offset = total % 8
									self._decoder.axes[next_axis].size = size
								next_axis = next_axis + 1
								if next_axis < AXIS_COUNT:
									next_axis = AxisType(next_axis)
							total += size
					elif kind == GenericDesktopPage.Hatswitch:
						if count * size != 4:
							raise UnparsableDescriptor("Invalid size for Hatswitch (%sb)" % (count * size,))
						if next_axis + 1 < AXIS_COUNT:
							log.debug("Found hat #%s at bit %s", int(next_axis), total)
							if config:
								target, axis_data = self._build_axis_maping(next_axis, config, AxisMode.HATSWITCH)
								if axis_data:
									axis_data.byte_offset = total // 8
									axis_data.bit_offset = total % 8
									self._decoder.axes[target] = axis_data
							else:
								self._decoder.axes[next_axis] = AxisData(mode=AxisMode.HATSWITCH)
								self._decoder.axes[next_axis].byte_offset = total // 8
								self._decoder.axes[next_axis].bit_offset = total % 8
								self._decoder.axes[next_axis].data.hatswitch.min = STICK_PAD_MIN
								self._decoder.axes[next_axis].data.hatswitch.max = STICK_PAD_MAX
							# Hatswitch is little special as it covers 2 axes at once
							next_axis = next_axis + 2
							if next_axis < AXIS_COUNT:
								next_axis = AxisType(next_axis)
						total += 4
					elif kind == UsagePage.ButtonPage:
						if self._decoder.buttons.enabled:
							raise UnparsableDescriptor("HID descriptor with two sets of buttons")
						if count * size < 8:
							buttons_size = 8
						elif count * size < 32:
							buttons_size = 32
						else:
							raise UnparsableDescriptor("Too many buttons (up to 32 supported)")
						log.debug("Found %s buttons at bit %s", count, total)
						self._decoder.buttons = ButtonData(
							enabled=True,
							byte_offset=total // 8,
							bit_offset=total % 8,
							size=buttons_size,
							button_count=count,
							button_map=self._build_button_map(config),
						)
						total += count * size
					else:
						log.debug("Skipped over %s bits for %s at bit %s", count * size, kind, total)
						total += count * size

		self._decoder.packet_size = total // 8
		if total % 8 > 0:
			self._decoder.packet_size += 1
		self._decoder.packet_size = min(self._decoder.packet_size, max_size)
		log.debug("Packet size: %s", self._decoder.packet_size)

	@staticmethod
	def find_sys_devices_descriptor(vid: int, pid: int) -> str | None:
		"""Finds, loads and returns HID descriptor available somewhere deep in /sys/devices structure.

		Done by walking /sys/devices recursivelly, searching for file named
		'report_descriptor' in subdirectory with name contining vid and pid.

		This is very much prefered before loading HID descriptor from device,
		as some controllers are presenting descriptor that are completly
		broken and kernel already deals with it.
		"""

		def recursive_search(pattern: str, path: str) -> str | None:
			for name in os.listdir(path):
				full_path = os.path.join(path, name)
				if name == "report_descriptor" and pattern in os.path.split(path)[-1].lower():
					return full_path
				try:
					if os.path.islink(full_path):
						# Recursive stuff in /sys ftw...
						continue
					if os.path.isdir(full_path):
						r = recursive_search(pattern, full_path)
						if r:
							return r
				except OSError:
					pass
			return None

		pattern = ":%.4x:%.4x" % (vid, pid)
		full_path = recursive_search(pattern, SYS_DEVICES)
		try:
			if full_path:
				log.debug("Loading descriptor from '%s'", full_path)
				temp_list = None
				with open(full_path, "rb") as read_file:
					temp_list = [x for x in read_file.read(1024)]

				return temp_list
		except Exception as e:
			log.exception(e)
		return None

	def close(self) -> None:
		# Called when pad is disconnected
		USBDevice.close(self)
		if self._ready:
			self.daemon.remove_controller(self)
			self._ready = False

	def get_type(self) -> str:
		return "hid"

	def _generate_id(self) -> str:
		"""ID is generated as 'hid0000:1111' where first number is vendor and 2nd product id.

		If two or more controllers with same vendor/product IDs are added,
		':X' is added, where 'X' starts as 1 and increases as controllers with same ids are connected.
		"""
		magic_number = 1
		vid, pid = self.device.getVendorID(), self.device.getProductID()
		id = "hid%.4x:%.4x" % (vid, pid)
		while id in self.daemon.get_active_ids():
			id = "hid%.4x:%.4x:%s" % (vid, pid, magic_number)
			magic_number += 1
		return id

	def get_id(self) -> str:
		return self._id

	def get_gui_config_file(self) -> str:
		return self.config_file

	def __repr__(self) -> str:
		vid, pid = self.device.getVendorID(), self.device.getProductID()
		return "<HID %.4x%.4x>" % (vid, pid)

	def test_input(self, endpoint: int, data):
		if not _lib.decode(ctypes.byref(self._decoder), data):
			# Returns True if anything changed
			return
		# Note: This is quite slow, but good enough for test mode
		code = 0
		for attr, trash in self._decoder.state._fields_:
			if attr == "buttons":
				continue
			if getattr(self._decoder.state, attr) != getattr(self._decoder.old_state, attr):
				# print("Axis", code, getattr(self._decoder.state, attr))
				sys.stdout.flush()
			code += 1

		pressed = self._decoder.state.buttons & ~self._decoder.old_state.buttons
		released = self._decoder.old_state.buttons & ~self._decoder.state.buttons
		for j in range(self._decoder.buttons.button_count):
			mask = 1 << j
			if pressed & mask:
				print("ButtonPress", FIRST_BUTTON + j)
				sys.stdout.flush()
			if released & mask:
				print("ButtonRelease", FIRST_BUTTON + j)
				sys.stdout.flush()

	def input(self, endpoint: int, data):
		if _lib.decode(ctypes.byref(self._decoder), data):
			if self.mapper:
				self.mapper.input(self, self._decoder.old_state, self._decoder.state)

	def apply_config(self, config):
		# TODO: This?
		pass

	def disconnected(self):
		# TODO: This!
		pass

	# def configure(self, idle_timeout=None, enable_gyros=None, led_level=None):

	def set_led_level(self, level):
		# TODO: This?
		pass

	def set_gyro_enabled(self, enabled):
		# TODO: This, maybe.
		pass


class HIDDrv:
	def __init__(self, daemon: SCCDaemon):
		self.registered = set()
		self.config_files = {}
		self.configs = {}
		self.scan_files()
		self.daemon = daemon

	def hotplug_cb(self, device: USBDevice, handle: USBDeviceHandle) -> HIDController | None:
		vid, pid = device.getVendorID(), device.getProductID()
		if (vid, pid) in self.configs:
			return HIDController(device, self.daemon, handle, self.config_files[vid, pid], self.configs[vid, pid])

		return None

	def scan_files(self) -> None:
		"""Go through ~/.config/scc/devices and enable hotplug callback for every known HID device."""
		path = os.path.join(get_config_path(), "devices")
		if not os.path.exists(path):
			# Nothing to do
			return

		known = set()
		for name in os.listdir(path):
			if name.startswith("hid-") and name.endswith(".json"):
				vid, pid = name.split("-", 2)[1].split(":")[0:2]
				vid = int(vid, 16)
				pid = int(pid, 16)
				config_file = os.path.join(path, name)
				try:
					config = json.loads(open(config_file).read())
				except Exception:
					log.warning("Ignoring file that cannot be parsed: %s", name)
					continue

				self.config_files[vid, pid] = config_file
				self.configs[vid, pid] = config
				known.add((vid, pid))

		for new in known - self.registered:
			vid, pid = new
			register_hotplug_device(self.hotplug_cb, vid, pid)
			self.registered.add(new)

		for removed in self.registered - known:
			vid, pid = removed
			unregister_hotplug_device(self.hotplug_cb, vid, pid)
			self.registered.remove(removed)
			if (vid, pid) in self.config_files:
				del self.config_files[vid, pid]
			if (vid, pid) in self.configs:
				del self.configs[vid, pid]


def hiddrv_test(cls, args) -> int:
	"""Small input test used by GUI while setting up the device.

	Basically, if HID device works with this, it will work with daemon as well.
	"""
	from scc.device_monitor import create_device_monitor
	from scc.drivers.usb import _usb
	from scc.poller import Poller
	from scc.scripts import InvalidArguments

	try:
		if ":" in args[0]:
			args[0:1] = args[0].split(":")
		vid = int(args[0], 16)
		pid = int(args[1], 16)
	except Exception:
		raise InvalidArguments()

	class FakeDaemon:
		def __init__(self) -> None:
			self.poller = Poller()
			self.dev_monitor = create_device_monitor(self)
			self.exitcode = -1

		def get_device_monitor(self) -> DeviceMonitor:
			return self.dev_monitor

		def add_error(self, id, error: str) -> None:
			fake_daemon.exitcode = 2
			log.error(error)

		def remove_error(*a) -> None:
			pass

		def get_poller(self) -> Poller:
			return self.poller

	fake_daemon = FakeDaemon()

	def cb(device: USBDevice, handle: USBDeviceHandle):
		try:
			return cls(device, None, handle, None, None, test_mode=True)
		except NotHIDDevice:
			print("%.4x:%.4x is not a HID device" % (vid, pid), file=sys.stderr)
			fake_daemon.exitcode = 3
		except UnparsableDescriptor as e:
			print("Invalid or unparsable HID descriptor", str(e), file=sys.stderr)
			fake_daemon.exitcode = 4
		except Exception as e:
			print("Failed to open device:", str(e), file=sys.stderr)
			fake_daemon.exitcode = 2

	_usb.set_daemon(fake_daemon)
	register_hotplug_device(cb, vid, pid)
	fake_daemon.dev_monitor.start()
	_usb.start()
	fake_daemon.dev_monitor.rescan()

	if fake_daemon.exitcode < 0:
		print("Ready")
	sys.stdout.flush()
	while fake_daemon.exitcode < 0:
		fake_daemon.poller.poll()
		_usb.mainloop()

	return fake_daemon.exitcode


def init(daemon: SCCDaemon, config: dict) -> bool:
	"""Called from scc-daemon."""
	d = HIDDrv(daemon)
	daemon.add_on_rescan(d.scan_files)
	return True


if __name__ == "__main__":
	"""Called when executed as script."""
	from scc.tools import init_logging, set_logging_level

	init_logging()
	set_logging_level(True, True)
	sys.exit(hiddrv_test(HIDController, sys.argv[1:]))



# The MIT License (MIT)
#
# Copyright (c) 2015 Stany MARCEL <stanypub@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import annotations

import ctypes
import os
from ctypes import POINTER, byref, c_bool, c_int16, c_int32, c_uint16
from enum import IntEnum
from math import copysign, fmod, sqrt
from typing import TYPE_CHECKING

from usb1.libusb1 import timeval

from scc.cheader import defines
from scc.tools import find_library

if TYPE_CHECKING:
	from ctypes import CDLL, Array, _Pointer

UNPUT_MODULE_VERSION = 9

# Get All defines from linux headers
if os.path.exists("/usr/include/linux/input-event-codes.h"):
	CHEAD = defines("/usr/include", "linux/input-event-codes.h")
elif os.path.exists(os.path.split(__file__)[0] + "/input-event-codes.h"):
	CHEAD = defines(os.path.split(__file__)[0], "input-event-codes.h")
else:
	CHEAD = defines("/usr/include", "linux/input.h")

MAX_FEEDBACK_EFFECTS = 4


if TYPE_CHECKING:
	# Static analyzers cannot see enum members added through locals().update().
	# Regenerate this stub with scripts/generate-uinput-stubs.py when the
	# canonical Linux input header changes.
	from scc._uinput_codes import Axes, Keys, KeysOnly, Rels
else:
	class Keys(IntEnum):
		"""Keys enum contains all keys and button from linux/uinput.h (KEY_* BTN_*)."""

		locals().update({i: CHEAD[i] for i in CHEAD if i.startswith(("KEY_", "BTN_"))})


	class KeysOnly(IntEnum):
		"""Keys enum contains all keys from linux/uinput.h (KEY_*)."""

		locals().update({i: CHEAD[i] for i in CHEAD if i.startswith("KEY_")})


	class Axes(IntEnum):
		"""Axes enum contains all axes from linux/uinput.h (ABS_*)."""

		locals().update({i: CHEAD[i] for i in CHEAD if i.startswith("ABS_")})


	class Rels(IntEnum):
		"""Rels enum contains all rels from linux/uinput.h (REL_*)."""

		locals().update({i: CHEAD[i] for i in CHEAD if i.startswith("REL_")})


# Scan codes for each keys (taken from a logitech keyboard)
Scans = {
	Keys["KEY_ESC"]: 0x70029,
	Keys["KEY_F1"]: 0x7003A,
	Keys["KEY_F2"]: 0x7003B,
	Keys["KEY_F3"]: 0x7003C,
	Keys["KEY_F4"]: 0x7003D,
	Keys["KEY_F5"]: 0x7003E,
	Keys["KEY_F6"]: 0x7003F,
	Keys["KEY_F7"]: 0x70040,
	Keys["KEY_F8"]: 0x70041,
	Keys["KEY_F9"]: 0x70042,
	Keys["KEY_F10"]: 0x70043,
	Keys["KEY_F11"]: 0x70044,
	Keys["KEY_F12"]: 0x70045,
	Keys["KEY_SYSRQ"]: 0x70046,
	Keys["KEY_SCROLLLOCK"]: 0x70047,
	Keys["KEY_PAUSE"]: 0x70048,
	Keys["KEY_GRAVE"]: 0x70035,
	Keys["KEY_1"]: 0x7001E,
	Keys["KEY_2"]: 0x7001F,
	Keys["KEY_3"]: 0x70020,
	Keys["KEY_4"]: 0x70021,
	Keys["KEY_5"]: 0x70022,
	Keys["KEY_6"]: 0x70023,
	Keys["KEY_7"]: 0x70024,
	Keys["KEY_8"]: 0x70025,
	Keys["KEY_9"]: 0x70026,
	Keys["KEY_0"]: 0x70027,
	Keys["KEY_MINUS"]: 0x7002D,
	Keys["KEY_EQUAL"]: 0x7002E,
	Keys["KEY_BACKSPACE"]: 0x7002A,
	Keys["KEY_TAB"]: 0x7002B,
	Keys["KEY_Q"]: 0x70014,
	Keys["KEY_W"]: 0x7001A,
	Keys["KEY_E"]: 0x70008,
	Keys["KEY_R"]: 0x70015,
	Keys["KEY_T"]: 0x70017,
	Keys["KEY_Y"]: 0x7001C,
	Keys["KEY_U"]: 0x70018,
	Keys["KEY_I"]: 0x7000C,
	Keys["KEY_O"]: 0x70012,
	Keys["KEY_P"]: 0x70013,
	Keys["KEY_LEFTBRACE"]: 0x7002F,
	Keys["KEY_RIGHTBRACE"]: 0x70030,
	Keys["KEY_ENTER"]: 0x70028,
	Keys["KEY_CAPSLOCK"]: 0x70039,
	Keys["KEY_A"]: 0x70004,
	Keys["KEY_S"]: 0x70016,
	Keys["KEY_D"]: 0x70007,
	Keys["KEY_F"]: 0x70009,
	Keys["KEY_G"]: 0x7000A,
	Keys["KEY_H"]: 0x7000B,
	Keys["KEY_J"]: 0x7000D,
	Keys["KEY_K"]: 0x7000E,
	Keys["KEY_L"]: 0x7000F,
	Keys["KEY_SEMICOLON"]: 0x70033,
	Keys["KEY_APOSTROPHE"]: 0x70034,
	Keys["KEY_BACKSLASH"]: 0x70032,
	Keys["KEY_LEFTSHIFT"]: 0x700E1,
	Keys["KEY_102ND"]: 0x70064,
	Keys["KEY_Z"]: 0x7001D,
	Keys["KEY_X"]: 0x7001B,
	Keys["KEY_C"]: 0x70006,
	Keys["KEY_V"]: 0x70019,
	Keys["KEY_B"]: 0x70005,
	Keys["KEY_N"]: 0x70011,
	Keys["KEY_M"]: 0x70010,
	Keys["KEY_COMMA"]: 0x70036,
	Keys["KEY_DOT"]: 0x70037,
	Keys["KEY_SLASH"]: 0x70038,
	Keys["KEY_RIGHTSHIFT"]: 0x700E5,
	Keys["KEY_LEFTCTRL"]: 0x700E0,
	Keys["KEY_LEFTMETA"]: 0x700E3,
	Keys["KEY_LEFTALT"]: 0x700E2,
	Keys["KEY_SPACE"]: 0x7002C,
	Keys["KEY_RIGHTALT"]: 0x700E6,
	Keys["KEY_RIGHTMETA"]: 0x700E7,
	Keys["KEY_COMPOSE"]: 0x70065,
	Keys["KEY_RIGHTCTRL"]: 0x700E4,
	Keys["KEY_INSERT"]: 0x70049,
	Keys["KEY_HOME"]: 0x7004A,
	Keys["KEY_PAGEUP"]: 0x7004B,
	Keys["KEY_DELETE"]: 0x7004C,
	Keys["KEY_END"]: 0x7004D,
	Keys["KEY_PAGEDOWN"]: 0x7004E,
	Keys["KEY_UP"]: 0x70052,
	Keys["KEY_LEFT"]: 0x70050,
	Keys["KEY_DOWN"]: 0x70051,
	Keys["KEY_RIGHT"]: 0x7004F,
	Keys["KEY_NUMLOCK"]: 0x70053,
	Keys["KEY_KPSLASH"]: 0x70054,
	Keys["KEY_KPASTERISK"]: 0x70055,
	Keys["KEY_KPMINUS"]: 0x70056,
	Keys["KEY_KP7"]: 0x7005F,
	Keys["KEY_KP8"]: 0x70060,
	Keys["KEY_KP9"]: 0x70061,
	Keys["KEY_KPPLUS"]: 0x70057,
	Keys["KEY_KP4"]: 0x7005C,
	Keys["KEY_KP5"]: 0x7005D,
	Keys["KEY_KP6"]: 0x7005E,
	Keys["KEY_KP1"]: 0x70059,
	Keys["KEY_KP2"]: 0x7005A,
	Keys["KEY_KP3"]: 0x7005B,
	Keys["KEY_KPENTER"]: 0x70058,
	Keys["KEY_KP0"]: 0x70062,
	Keys["KEY_KPDOT"]: 0x70063,
	Keys["KEY_CONFIG"]: 0xC0183,
	Keys["KEY_PLAYPAUSE"]: 0xC00CD,
	Keys["KEY_MUTE"]: 0xC00E2,
	Keys["KEY_VOLUMEDOWN"]: 0xC00EA,
	Keys["KEY_VOLUMEUP"]: 0xC00E9,
	Keys["KEY_HOMEPAGE"]: 0xC0223,
	Keys["KEY_PREVIOUSSONG"]: 0xC00F0,
	Keys["KEY_NEXTSONG"]: 0xC00F1,
	Keys["KEY_BACK"]: 0xC00F2,
	Keys["KEY_FORWARD"]: 0xC00F3,
}


class InputEvent(ctypes.Structure):
	_fields_: list[tuple[str, type]] = [("time", timeval), ("type", c_uint16), ("code", c_uint16), ("value", c_int32)]


class FeedbackEvent(ctypes.Structure):
	_fields_: list[tuple[str, type]] = [
		("in_use", c_bool),
		("continuous_rumble", c_bool),
		("duration", c_int32),
		("delay", c_int32),
		("repetitions", c_int32),
		("type", c_uint16),
		("level", c_int16),
	]

	def __init__(self):
		self.in_use: bool = False


class UInput:
	"""UInput class permits to create a uinput device.

	See Gamepad, Mouse, Keyboard for examples
	"""

	def __init__(
		self,
		vendor: int,
		product: int,
		version: int,
		name: str,
		keys, # : Sequence[int],
		axes, # : Sequence[tuple[int]],
		rels, # : Sequence[int],
		keyboard: bool = False,
		rumble: bool = False,
	):
		self._k = keys
		self.name: str = name
		if not axes or len(axes) == 0:
			self._a, self._amin, self._amax, self._afuzz, self._aflat = [[]] * 5
		else:
			self._a, self._amin, self._amax, self._afuzz, self._aflat = zip(*axes)

		self._r = rels

		self._lib: CDLL = find_library("libuinput")
		self._ff_events: Array[_Pointer[FeedbackEvent]] | None = None
		if rumble:
			self._ff_events = (POINTER(FeedbackEvent) * MAX_FEEDBACK_EFFECTS)()
			for i in range(MAX_FEEDBACK_EFFECTS):
				self._ff_events[i].contents = FeedbackEvent()

		try:
			if self._lib.uinput_module_version() != UNPUT_MODULE_VERSION:
				raise Exception
		except Exception:
			import sys

			print("Invalid native module version. Please, recompile 'libuinput.so'", file=sys.stderr)
			print(
				"If you are running sc-controller from source, you can do this by removing 'build' directory",
				file=sys.stderr,
			)
			print("and running 'python setup.py build' or 'run.sh' script", file=sys.stderr)
			raise Exception("Invalid native module version")

		c_k = (ctypes.c_uint16 * len(self._k))(*self._k)
		c_a = (ctypes.c_uint16 * len(self._a))(*self._a)
		c_amin = (ctypes.c_int32 * len(self._amin))(*self._amin)
		c_amax = (ctypes.c_int32 * len(self._amax))(*self._amax)
		c_afuzz = (ctypes.c_int32 * len(self._afuzz))(*self._afuzz)
		c_aflat = (ctypes.c_int32 * len(self._aflat))(*self._aflat)
		c_r = (ctypes.c_uint16 * len(self._r))(*self._r)
		c_vendor = ctypes.c_uint16(vendor)
		c_product = ctypes.c_uint16(product)
		c_version = ctypes.c_uint16(version)
		c_keyboard = ctypes.c_int(keyboard)
		c_rumble = ctypes.c_int(MAX_FEEDBACK_EFFECTS if rumble else 0)
		c_name = ctypes.c_char_p(name.encode("utf-8") if type(name) is str else name)

		self._fd = self._lib.uinput_init(
			ctypes.c_int(len(self._k)),
			c_k,
			ctypes.c_int(len(self._a)),
			c_a,
			c_amin,
			c_amax,
			c_afuzz,
			c_aflat,
			ctypes.c_int(len(self._r)),
			c_r,
			c_keyboard,
			c_vendor,
			c_product,
			c_version,
			c_rumble,
			c_name,
		)
		if self._fd < 0:
			raise CannotCreateUInputException("Failed to create uinput device. Error code: %s" % (self._fd,))

	def getDescriptor(self):
		return self._fd

	def keyEvent(self, key: int, val: int):
		"""Generate a key or btn event.

		@param int axis		 key or btn event (KEY_* or BTN_*)
		@param int val		  event value
		"""
		self._lib.uinput_key(self._fd, ctypes.c_uint16(key), ctypes.c_int32(val))

	def axisEvent(self, axis: int, val: int):
		"""Generate a abs event (joystick/pad axes).

		@param int axis		 abs event (ABS_*)
		@param int val		  event value
		"""
		self._lib.uinput_abs(self._fd, ctypes.c_uint16(axis), ctypes.c_int32(val))

	def relEvent(self, rel: int, val: int):
		"""Generate a rel event (move move).

		@param int rel		  rel event (REL_*)
		@param int val		  event value
		"""
		self._lib.uinput_rel(self._fd, ctypes.c_uint16(rel), ctypes.c_int32(val))

	def scanEvent(self, val: int):
		"""Generate a scan event (MSC_SCAN).

		@param int val		  scan event value (scancode)
		"""
		self._lib.uinput_scan(self._fd, ctypes.c_int32(val))

	def synEvent(self):
		"""Generate a syn event."""
		self._lib.uinput_syn(self._fd)

	def setDelayPeriod(self, delay: int, period: int):
		"""Update delay period values for keyboard.

		@param int delay		delay in ms
		@param int period	   period is ms
		"""
		self._lib.uinput_set_delay_period(self._fd, ctypes.c_int32(delay), ctypes.c_int32(period))

	def keyManaged(self, ev):
		return ev in self._k

	def axisManaged(self, ev):
		return ev in self._a

	def relManaged(self, ev):
		return ev in self._r

	def ff_read(self):
		"""Return effect that should be played or None if there were no such request."""
		if self._ff_events:
			id = self._lib.uinput_ff_read(self._fd, MAX_FEEDBACK_EFFECTS, byref(self._ff_events))
			if id >= 0:
				return self._ff_events[id].contents
		return None

	def __del__(self):
		if self._lib:
			self._lib.uinput_destroy(self._fd)


class Gamepad(UInput):
	"""Gamepad uinput class, create a Xbox360 gamepad device."""

	def __init__(self, name):
		super().__init__(
			vendor=0x045E,
			product=0x028E,
			version=1,
			name=name,
			keys=[
				Keys["BTN_START"],
				Keys["BTN_MODE"],
				Keys["BTN_SELECT"],
				Keys["BTN_A"],
				Keys["BTN_B"],
				Keys["BTN_X"],
				Keys["BTN_Y"],
				Keys["BTN_TL"],
				Keys["BTN_TR"],
				Keys["BTN_THUMBL"],
				Keys["BTN_THUMBR"],
			],
			axes=[
				(Axes["ABS_X"], -32768, 32767, 16, 128),
				(Axes["ABS_Y"], -32768, 32767, 16, 128),
				(Axes["ABS_RX"], -32768, 32767, 16, 128),
				(Axes["ABS_RY"], -32768, 32767, 16, 128),
				(Axes["ABS_Z"], 0, 255, 0, 0),
				(Axes["ABS_RZ"], 0, 255, 0, 0),
				(Axes["ABS_HAT0X"], -1, 1, 0, 0),
				(Axes["ABS_HAT0Y"], -1, 1, 0, 0),
			],
			rels=[],
		)


class Mouse(UInput):
	"""Mouse uinput class, create a mouse device.

	moveEvent can emulate free ball rotation of a track ball
	updateParams permit to upgrade ball model and move scale
	"""

	DEFAULT_XSCALE = 0.006
	DEFAULT_YSCALE = 0.006

	DEFAULT_SCR_XSCALE = 0.0005
	DEFAULT_SCR_YSCALE = 0.0005

	def __init__(self, name):
		super().__init__(
			vendor=0x28DE,
			product=0x1142,
			version=1,
			name=name,
			keys=[Keys.BTN_LEFT, Keys.BTN_RIGHT, Keys.BTN_MIDDLE, Keys.BTN_SIDE, Keys.BTN_EXTRA],
			axes=[],
			rels=[Rels.REL_X, Rels.REL_Y, Rels.REL_WHEEL, Rels.REL_HWHEEL],
		)
		self.updateParams()
		self.updateScrollParams()
		self.reset()

	def reset(self):
		"""Reset internal counters, especially the one used for wheel.

		Fixes scroll wheel feedback desynchronisation, as reported
		in https://github.com/kozec/sc-controller/issues/222
		"""
		self._scr_dx = 0.0
		self._scr_dy = 0.0
		self._dx = 0.0
		self._dy = 0.0

	def updateParams(self, xscale: float = DEFAULT_XSCALE, yscale: float = DEFAULT_YSCALE):
		"""Update Movement parameters.

		@param float mass	   mass in g of the ball
		@param float r		  radius in m of the ball
		@param int ampli		integer amplitude for move from border to border
		@param float degree	 degree of rotation of the ball for move from border to border
		@param float xscale	 scale applied on move param to input event on x axis
		@param float yscale	 scale applied on move param to input event on y axis
		"""
		self._xscale = xscale
		self._yscale = yscale

	def updateScrollParams(self, xscale: float = DEFAULT_SCR_XSCALE, yscale: float = DEFAULT_SCR_YSCALE):
		"""Update Scroll parameters

		@param float mass	   mass in g of the ball
		@param float r		  radius in m of the ball
		@param float friction   constat friction force applied to the ball
		@param int ampli		integer amplitude for move from border to border
		@param float degree	 degree of rotation of the ball for move from border to border
		@param float xscale	 scale applied on move param to input event on x axis
		@param float yscale	 scale applied on move param to input event on y axis
		"""
		self._scr_xscale = xscale
		self._scr_yscale = yscale

	def moveEvent(self, dx: int = 0, dy: int = 0, time_elapsed: float = 0.0):
		"""Generate move events from parametters and displacement.

		@param int dx		   delta movement from last call on x axis
		@param int dy		   delta movement from last call on y axis
		"""
		_syn = False

		# Clear mouse axis remainders if axis direction has changed
		if dx == 0 or ((dx > 0) != (self._dx > 0)):
			self._dx = 0

		# Clear mouse axis remainders if axis direction has changed
		if dy == 0 or ((dy > 0) != (self._dy > 0)):
			self._dy = 0

		# Base speed around 8 ms standard
		# (base USB poll rate for Steam Controller)
		baseFactor = time_elapsed * 125.0
		self._dx += dx * self._xscale * baseFactor
		self._dy += dy * self._yscale * baseFactor
		# self._factorDeadzone(dx, dy, time_elapsed)

		if int(self._dx):
			self._dx = self._dx - (fmod(self._dx * 100.0, 1.0) / 100.0)
			self.relEvent(rel=Rels.REL_X, val=int(self._dx))
			self._dx -= int(self._dx)
			_syn = True
		if int(self._dy):
			self._dy = self._dy - (fmod(self._dy * 100.0, 1.0) / 100.0)
			self.relEvent(rel=Rels.REL_Y, val=int(self._dy))
			self._dy -= int(self._dy)
			_syn = True
		if _syn:
			self.synEvent()

	def moveStickEvent(self, dx: float = 0.0, dy: float = 0.0, time_elapsed: float = 0.0):
		"""Generate move events from parametters and displacement.

		@param float dx		   delta movement from last call on x axis
		@param float dy		   delta movement from last call on y axis
		"""
		_syn = False

		# Clear mouse axis remainders if axis direction has changed
		if dx == 0 or ((dx > 0) != (self._dx > 0)):
			self._dx = 0

		# Clear mouse axis remainders if axis direction has changed
		if dy == 0 or ((dy > 0) != (self._dy > 0)):
			self._dy = 0

		self._dx += dx
		self._dy += dy
		# self._factorDeadzone(dx, dy, time_elapsed)

		if int(self._dx):
			self._dx = self._dx - (fmod(self._dx * 100.0, 1.0) / 100.0)
			self.relEvent(rel=Rels.REL_X, val=int(self._dx))
			self._dx -= int(self._dx)
			_syn = True
		if int(self._dy):
			self._dy = self._dy - (fmod(self._dy * 100.0, 1.0) / 100.0)
			self.relEvent(rel=Rels.REL_Y, val=int(self._dy))
			self._dy -= int(self._dy)
			_syn = True
		if _syn:
			self.synEvent()

	def clearRemainders(self):
		self._dx = 0
		self._dy = 0

	def _factorDeadzone(self, dx: int, dy: int, time_elapsed: float):
		"""Take raw event and adjust based on assigned dead zone setting.

		@param int dx				delta movement from last call on x axis
		@param int dy				delta movement from last call on y axis
		@param float time_elapsed		time elapsed in sec.
		"""
		# print("COMING IN {} {}".format(dx, dy))
		# deadzonetmp = 15
		deadzonetmp = 22
		# offset = 0.297477440456902
		offset = 0.375  # 0.8 #0.6 #0.45

		# if (dx == 0 or ((dx > 0) != (self._dx > 0))):
		# self._dx = 0

		# if (dy == 0 or ((dy > 0) != (self._dy > 0))):
		# self._dy = 0

		_hyp = sqrt((dx**2) + (dy**2))
		unitx = 0
		unity = 0
		deadzoneX = deadzonetmp
		deadzoneY = deadzonetmp
		if _hyp != 0.0:
			unitx = dx / _hyp
			unity = dy / _hyp
			deadzoneX = int(deadzonetmp * unitx)
			deadzoneY = int(deadzonetmp * unity)

		if abs(dx) > abs(deadzoneX):
			beforedx = dx
			dx -= copysign(deadzoneX, dx)
			# print("DX: {} {} {} {}".format(beforedx, dx, deadzoneX, unitx))
		else:
			dx = 0

		if abs(dy) > abs(deadzoneY):
			beforedy = dy
			dy -= copysign(deadzoneY, dy)
			# print("DY: {} {} {} {}".format(beforedy, dy, deadzoneY, unity))
		else:
			dy = 0

		# throttla = 1.43
		throttla = 1.428
		offman = 28  # 30
		tempx = 0.0
		tempy = 0.0
		# Throttle low end of X axis movement
		if dx != 0.0:
			if abs(dx) < (abs(unitx) * offman):
				signx = copysign(1.0, dx)
				ratioX = abs(dx) / offman
				# print("OLD {} | NEW {}".format(tempx, tempx ** 1.5))
				dx = ratioX**throttla * signx * offman

		if dx != 0.0:
			tempx = dx * (time_elapsed * 125.0) * self._xscale + (abs(unitx) * copysign(offset, dx))
			self._dx += tempx
		else:
			# print("UP IN HERE {}".format(self._dx))
			self._dx = 0

		# Throttle low end of Y axis movement
		if dy != 0.0:
			if abs(dy) < (abs(unity) * offman):
				signy = copysign(1.0, dy)
				ratioY = abs(dy) / offman
				# print("OLD {} | NEW {}".format(tempy, tempy ** 1.5))
				dy = ratioY**throttla * signy * offman

		if dy != 0.0:
			tempy = dy * (time_elapsed * 125.0) * self._yscale + (abs(unity) * copysign(offset, dy))
			self._dy += tempy
		else:
			self._dy = 0

	def scrollEvent(self, dx: int = 0, dy: int = 0):
		"""Generate scroll events from parametters and displacement.

		@param int dx		   delta movement from last call on x axis
		@param int dy		   delta movement from last call on y axis

		@return float		   absolute distance moved this tick # TODO <- Does not return????
		"""
		# Compute mouse mouvement from interger part of d * scale
		self._scr_dx += dx * self._scr_xscale
		self._scr_dy += dy * self._scr_yscale
		_syn = False
		if int(self._scr_dx):
			self.relEvent(rel=Rels.REL_HWHEEL, val=int(copysign(1, self._scr_dx)))
			self._scr_dx -= int(self._scr_dx)
			_syn = True
		if int(self._scr_dy):
			self.relEvent(rel=Rels.REL_WHEEL, val=int(copysign(1, self._scr_dy)))
			self._scr_dy -= int(self._scr_dy)
			_syn = True
		if _syn:
			self.synEvent()


class Keyboard(UInput):
	"""Keyboard uinput class, create a keyboard device.

	pressEvent permit to generate a key pressed and with scan events
	releaseEvent permit to generate a key released and with scan events

	autorepead delay and period are preset respectively to 250ms and 33ms
	setDelayPeriod permits to update these values
	"""

	def __init__(self, name):
		super().__init__(
			vendor=0x28DE, product=0x1142, version=1, name=name, keys=Scans.keys(), axes=[], rels=[], keyboard=True,
		)
		self.setDelayPeriod(250, 33)
		self._dx = 0.0
		self._pressed = set()

	def pressEvent(self, keys: list):
		"""Generate key press event with corresponding scan codes.

		Events are generated only for new keys.

		@param list of Keys keys		keys to press
		"""
		new = [k for k in keys if k not in self._pressed]
		for i in new:
			self.scanEvent(Scans[i])
			self.keyEvent(i, 1)
		if new:
			self.synEvent()
			self._pressed |= set(new)

	def releaseEvent(self, keys: list | None = None):
		"""Generate key release event with corresponding scan codes.

		Events are generated only for keys that was pressed

		@param list of Keys keys		keys to release, give None or empty list
										to release all
		"""
		if keys and len(keys):
			rem = [k for k in keys if k in self._pressed]
		else:
			rem = list(self._pressed)
		for i in rem:
			self.scanEvent(Scans[i])
			self.keyEvent(i, 0)
		if rem:
			self.synEvent()
			self._pressed -= set(rem)


class Dummy:
	"""Fake uinput device that does nothing, but has all required methods."""

	def __init__(self, *a, **b):
		pass

	def keyEvent(self, *a, **b):
		pass

	axisEvent = keyEvent
	relEvent = keyEvent
	scanEvent = keyEvent
	synEvent = keyEvent
	setDelayPeriod = keyEvent
	updateParams = keyEvent
	updateScrollParams = keyEvent
	moveEvent = keyEvent
	scrollEvent = keyEvent
	pressEvent = keyEvent
	releaseEvent = keyEvent
	reset = keyEvent

	def keyManaged(self, ev):
		return False

	axisManaged = keyManaged
	relManaged = keyManaged


class CannotCreateUInputException(Exception):
	# Special case when message should be displayed in UI
	pass



"""SC Controller - DualSense driver

Extends HID driver with DS5-specific options.
"""

import ctypes
import logging
import math
import os
import sys
import time
import zlib
from enum import IntEnum

from scc.constants import (
	STICK_PAD_MAX,
	STICK_PAD_MIN,
	STICK_PAD_RES,
	ControllerFlags,
	HapticPos,
	SCButtons,
)
from scc.controller import Controller
from scc.drivers.evdevdrv import (
	HAVE_EVDEV,
	EvdevController,
	get_axes,
	get_evdev_devices_from_syspath,
	make_new_device,
)
from scc.drivers.hiddrv import (
	BUTTON_COUNT,
	AxisData,
	AxisDataUnion,
	AxisMode,
	AxisModeData,
	AxisType,
	ButtonData,
	HatswitchModeData,
	HIDController,
	HIDDecoder,
	_lib,
	button_to_bit,
	hiddrv_test,
)
from scc.drivers.usb import register_hotplug_device
from scc.lib.hidraw import HIDRaw
from scc.sccdaemon import SCCDaemon
from scc.tools import init_logging, set_logging_level

log = logging.getLogger("DS5")

VENDOR_ID = 0x054C
PRODUCT_ID = 0x0CE6
# TODO(Martin): We do not currently support the 4 extra buttons
EDGE_PRODUCT_ID = 0x0DF2
PRODUCT_IDS = (PRODUCT_ID, EDGE_PRODUCT_ID)


OPERATING_MODE_DS5_BT = 0x31


class OperatingMode(IntEnum):
	DS4_COMPATIBILITY_MODE = 1 << 0
	DS5_MODE = 1 << 1
	DS5_MODE_BT = 1 << 5 | 1 << 4 | 1 << 0


class PhysicalEffectControl(IntEnum):
	ENABLE_HAPTICS = 1 << 0 | 1 << 1
	TRIGGER_EFFECTS_RIGHT = 1 << 2
	TRIGGER_EFFECTS_LEFT = 1 << 3


class LightEffectControl(IntEnum):
	MIC_MUTE_LED_CONTROL_ENABLE = 1 << 0
	POWER_SAVE_CONTROL_ENABLE = 1 << 1
	LIGHTBAR_CONTROL_ENABLE = 1 << 2
	RELEASE_LEDS = 1 << 3
	PLAYER_INDICATOR_CONTROL_ENABLE = 1 << 4


class DualSenseHIDOutput(ctypes.Structure):
	_fields_ = [
		("operating_mode", ctypes.c_ubyte),
		("physical_effect_control", ctypes.c_ubyte),
		("light_effect_control", ctypes.c_ubyte),
		("motor_right", ctypes.c_ubyte),
		("motor_left", ctypes.c_ubyte),
		("unknown2", ctypes.c_ubyte * 4),
		("mute_button_led", ctypes.c_ubyte),
		("power_save_control", ctypes.c_ubyte),
		("right_trigger_effect", ctypes.c_ubyte * 11),
		("left_trigger_effect", ctypes.c_ubyte * 11),
		("unknown3", ctypes.c_ubyte * 8),
		("lightbar_control", ctypes.c_ubyte),
		("lightbar_setup", ctypes.c_ubyte),
		("led_brightness", ctypes.c_ubyte),
		("player_leds", ctypes.c_ubyte),
		("lightbar_red", ctypes.c_ubyte),
		("lightbar_green", ctypes.c_ubyte),
		("lightbar_blue", ctypes.c_ubyte),
	]


class DualSenseHIDOutputBT(ctypes.Structure):
	_fields_ = [
		("operating_mode", ctypes.c_byte),
		("data_id_byte", ctypes.c_byte),
		("physical_effect_control", ctypes.c_byte),
		("light_effect_control", ctypes.c_byte),
		("motor_right", ctypes.c_byte),
		("motor_left", ctypes.c_byte),
		("unknown2", ctypes.c_byte * 4),
		("mute_button_led", ctypes.c_byte),
		("power_save_control", ctypes.c_byte),
		("right_trigger_effect", ctypes.c_byte * 11),
		("left_trigger_effect", ctypes.c_byte * 11),
		("unknown3", ctypes.c_byte * 6),
		("brightlite", ctypes.c_byte),
		("unknown6", ctypes.c_byte * 2),
		("lightbar_control", ctypes.c_byte),
		# ('lightbar_setup', ctypes.c_byte),
		("led_brightness", ctypes.c_byte),
		("player_leds", ctypes.c_byte),
		("lightbar_red", ctypes.c_byte),
		("lightbar_green", ctypes.c_byte),
		("lightbar_blue", ctypes.c_byte),
		("unknown4", ctypes.c_byte * 25),
		# ('unknown5', ctypes.c_byte * 4),
		("crc32", ctypes.c_byte * 4),
		# ('crc32', ctypes.c_uint16),
	]


class DualSenseHIDInputBT(ctypes.Structure):
	_fields_ = [
		("report_id", ctypes.c_byte),
		("unknown1", ctypes.c_byte),
		("lx", ctypes.c_ubyte),
		("ly", ctypes.c_ubyte),
	]


class DualSenseBTControllerInput(ctypes.Structure):
	_fields_ = [
		("type", ctypes.c_uint16),
		("buttons", ctypes.c_uint32),
		("ltrig", ctypes.c_uint8),
		("rtrig", ctypes.c_uint8),
		("stick_x", ctypes.c_int32),
		("stick_y", ctypes.c_int32),
		("lpad_x", ctypes.c_int32),
		("lpad_y", ctypes.c_int32),
		("rpad_x", ctypes.c_int32),
		("rpad_y", ctypes.c_int32),
		("accel_x", ctypes.c_int32),
		("accel_y", ctypes.c_int32),
		("accel_z", ctypes.c_int32),
		("gpitch", ctypes.c_int32),
		("groll", ctypes.c_int32),
		("gyaw", ctypes.c_int32),
		("q1", ctypes.c_int32),
		("q2", ctypes.c_int32),
		("q3", ctypes.c_int32),
		("q4", ctypes.c_int32),
		("cpad_x", ctypes.c_uint16),
		("cpad_y", ctypes.c_uint16),
	]


ICON_COLORS = [
	(0.0, 1.0, 0.0),  # 0
	(0.0, 0.0, 1.0),  # 1
	(1.0, 0.0, 0.0),  # 2
	(1.0, 1.0, 0.0),  # 3
	(0.0, 1.0, 1.0),  # 4
	(1.0, 0.4, 0.0),  # 5
	(1.0, 0.0, 1.0),  # 6
]


class DS5Controller(HIDController):
	# Most of axes are the same
	BUTTON_MAP = (
		SCButtons.X,
		SCButtons.A,
		SCButtons.B,
		SCButtons.Y,
		SCButtons.LB,
		SCButtons.RB,
		1 << 64,
		1 << 64,
		SCButtons.BACK,
		SCButtons.START,
		SCButtons.STICKPRESS,
		SCButtons.RPAD,
		SCButtons.C,
		SCButtons.CPADPRESS,
	)

	flags = (
		ControllerFlags.EUREL_GYROS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.HAS_CPAD
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.NO_GRIPS
	)

	def __init__(self, device, daemon, handle, config_file, config, test_mode=False):
		self._outputs = {}
		self._feedback_output = DualSenseHIDOutput(
			operating_mode=OperatingMode.DS5_MODE,
			physical_effect_control=PhysicalEffectControl.ENABLE_HAPTICS,
			motor_left=0,
			motor_right=0,
		)
		self._feedback_cancel_task = None
		super().__init__(device, daemon, handle, config_file, config, test_mode)

	def _load_hid_descriptor(self, config, max_size, vid, pid, test_mode):
		# Overrided and hardcoded
		self._decoder = HIDDecoder()

		# Dpad works on DualSense!
		self._decoder.axes[AxisType.AXIS_LPAD_X] = AxisData(
			mode=AxisMode.HATSWITCH,
			byte_offset=8,
			size=8,
			data=AxisDataUnion(
				hatswitch=HatswitchModeData(
					button=SCButtons.LPAD | SCButtons.LPADTOUCH, min=STICK_PAD_MIN, max=STICK_PAD_MAX,
				),
			),
		)

		# Sticks are the same as DS4
		self._decoder.axes[AxisType.AXIS_STICK_X] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=1,
			size=8,
			data=AxisDataUnion(axis=AxisModeData(scale=1.0, offset=-127.5, clamp_max=257, deadzone=2)),
		)
		self._decoder.axes[AxisType.AXIS_STICK_Y] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=2,
			size=8,
			data=AxisDataUnion(axis=AxisModeData(scale=-1.0, offset=127.5, clamp_max=257, deadzone=2)),
		)
		self._decoder.axes[AxisType.AXIS_RPAD_X] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=3,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(button=SCButtons.RPADTOUCH, scale=1.0, offset=-127.5, clamp_max=257, deadzone=2),
			),
		)
		self._decoder.axes[AxisType.AXIS_RPAD_Y] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=4,
			size=8,
			data=AxisDataUnion(
				axis=AxisModeData(button=SCButtons.RPADTOUCH, scale=-1.0, offset=127.5, clamp_max=257, deadzone=2),
			),
		)

		# Triggers
		self._decoder.axes[AxisType.AXIS_LTRIG] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=5,
			size=8,  # Not sure about the size
			data=AxisDataUnion(axis=AxisModeData(scale=1.0, clamp_max=1, deadzone=10)),
		)
		self._decoder.axes[AxisType.AXIS_RTRIG] = AxisData(
			mode=AxisMode.AXIS,
			byte_offset=6,
			size=8,  # Not sure about the size
			data=AxisDataUnion(axis=AxisModeData(scale=1.0, clamp_max=1, deadzone=10)),
		)

		# Gyro
		# Leaving the AxisMode naming to match DS4
		self._decoder.axes[AxisType.AXIS_GPITCH] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=16)  # Pitch found
		self._decoder.axes[AxisType.AXIS_GROLL] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=20)  # Roll
		self._decoder.axes[AxisType.AXIS_GYAW] = AxisData(mode=AxisMode.DS4ACCEL, byte_offset=18)  # Yaw found
		self._decoder.axes[AxisType.AXIS_Q1] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=26)
		self._decoder.axes[AxisType.AXIS_Q2] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=22)
		self._decoder.axes[AxisType.AXIS_Q3] = AxisData(mode=AxisMode.DS4GYRO, byte_offset=24)

		# Touchpad
		self._decoder.axes[AxisType.AXIS_CPAD_X] = AxisData(mode=AxisMode.DS4TOUCHPAD, byte_offset=34)  # DualSense X
		self._decoder.axes[AxisType.AXIS_CPAD_Y] = AxisData(
			mode=AxisMode.DS4TOUCHPAD, byte_offset=35, bit_offset=4,
		)  # DualSense Y

		# Button maps seem to work for standard arrangement (matching Xbox360)
		# Not enough information about the button event triggered when LT && RT are pressed?
		# Could be connected to adaptive triggers?
		self._decoder.buttons = ButtonData(
			enabled=True,
			byte_offset=8,
			bit_offset=4,
			size=14,  # Not sure about bit offset
			button_count=14,
		)

		if test_mode:
			for x in range(BUTTON_COUNT):
				self._decoder.buttons.button_map[x] = x
		else:
			for x in range(BUTTON_COUNT):
				self._decoder.buttons.button_map[x] = 64
			for x, sc in enumerate(DS5Controller.BUTTON_MAP):
				self._decoder.buttons.button_map[x] = button_to_bit(sc)

		self._packet_size = 64

	def input(self, endpoint: int, data: bytearray) -> None:
		# Special override for CPAD touch button
		if _lib.decode(ctypes.byref(self._decoder), bytes(data)):
			if self.mapper:
				if data[33] >> 7:
					# cpad is not touched
					self._decoder.state.buttons &= ~SCButtons.CPADTOUCH
				else:
					self._decoder.state.buttons |= SCButtons.CPADTOUCH
				self.mapper.input(self, self._decoder.old_state, self._decoder.state)

	def feedback(self, data):
		position, amplitude, period, count = data.data

		normalized_amp = float(amplitude) / 0x8000
		clamped_amp = int(normalized_amp * 0xFF)
		half_amp = int(normalized_amp * 0x80)

		if position == HapticPos.LEFT:
			# NOTE: the left motor is heavier, so we must give it less oomph
			self._feedback_output.motor_left = half_amp
		elif position == HapticPos.RIGHT:
			self._feedback_output.motor_right = clamped_amp
		elif position == HapticPos.BOTH:
			self._feedback_output.motor_right = clamped_amp
			self._feedback_output.motor_left = half_amp

		duration = float(period) * count / 0x10000
		# The motors don't seem to perform reliably when shut off under 20ms
		duration = max(duration, 0.02)

		self.schedule_output("feedback", self._feedback_output)

		def clear_feedback(mapper):
			self._feedback_output.motor_right = self._feedback_output.motor_left = 0
			self.schedule_output("feedback", self._feedback_output)

		if self._feedback_cancel_task:
			self._feedback_cancel_task.cancel()
		self._feedback_cancel_task = self.mapper.schedule(duration, clear_feedback)

	def apply_config(self, config):
		icon = config["icon"]
		led_level = config["led_level"]
		self.configure(icon=icon, led_level=led_level)

	def configure(self, icon=None, led_level=100):
		lightbar_color = (0.0, 0.0, 1.0)  # blue by default
		if icon:
			basename, ext = icon.rsplit(".", 1)
			parts = basename.rsplit("-", 1)
			if parts:
				raw_idx = parts[-1]
				try:
					icon_idx = int(raw_idx)
				except ValueError:
					pass
				else:
					if icon_idx < len(ICON_COLORS):
						lightbar_color = ICON_COLORS[icon_idx]

		led_level_norm = float(led_level) / 100
		lightbar_color_bytes = tuple(int(color_norm * led_level_norm * 255) for color_norm in lightbar_color)

		output = DualSenseHIDOutput(
			operating_mode=OperatingMode.DS5_MODE,
			light_effect_control=LightEffectControl.LIGHTBAR_CONTROL_ENABLE,
			lightbar_red=lightbar_color_bytes[0],
			lightbar_green=lightbar_color_bytes[1],
			lightbar_blue=lightbar_color_bytes[2],
		)
		self.schedule_output("lightbar", output)

	def get_gyro_enabled(self):
		# Cannot be actually turned off, so it's always active
		# TODO: Maybe emulate turning off?
		return True

	def get_type(self):
		return "ds5"

	def get_gui_config_file(self):
		return "ds5-config.json"

	def __repr__(self):
		return "<DS5Controller %s>" % (self.get_id(),)

	def _generate_id(self):
		"""ID is generated as 'ds5' or 'ds5:X' where 'X' starts as 1 and increases as controllers with same ids are connected."""
		magic_number = 1
		id = "ds5"
		while id in self.daemon.get_active_ids():
			id = f"ds5:{magic_number}"
			magic_number += 1
		return id

	def schedule_output(self, output_id, output):
		self._outputs[output_id] = output

	def flush(self):
		super().flush()

		while self._outputs:
			output_id, output = self._outputs.popitem()
			data = bytes(bytearray(output).ljust(64, b"\x00"))
			self.handle.interruptWrite(3, data)


class DS5HidRawDriver:
	def __init__(self, daemon: SCCDaemon, config: dict):
		self.config = config
		self.daemon = daemon
		for product_id in PRODUCT_IDS:
			daemon.get_device_monitor().add_callback(
				"bluetooth", VENDOR_ID, product_id, self.make_bt_hidraw_callback, None,
			)

	def retry(self, syspath: str):
		pass

	def make_bt_hidraw_callback(self, syspath: str, *whatever):
		hidrawname = self.daemon.get_device_monitor().get_hidraw(syspath)
		if hidrawname is None:
			return None

		# log.debug(whatever)
		try:
			dev = HIDRaw(open(os.path.join("/dev/", hidrawname), "w+b"))
			return DS5HidRawController(self, syspath, dev)
		except Exception as e:
			log.exception(e)
			return None


class DS5HidRawController(Controller):
	class _DPadOutputValues:
		def __init__(self, x, y):
			self.x = x
			self.y = y

	"""# (x, y) values
	DPAD_STATE_TYPES = {
		0: (0, STICK_PAD_MAX), # Up
		1: (STICK_PAD_MAX, STICK_PAD_MAX), # UpRight
		2: (STICK_PAD_MAX, 0), # Right
		3: (STICK_PAD_MAX, STICK_PAD_MIN), # DownRight
		4: (0, STICK_PAD_MIN), # Down
		5: (STICK_PAD_MIN, STICK_PAD_MIN), # DownLeft
		6: (STICK_PAD_MIN, 0), # Left
		7: (STICK_PAD_MIN, STICK_PAD_MAX), # UpLeft
		8: (0, 0) # Centered
	}
	"""

	# (x, y) values
	DPAD_STATE_TYPES = {
		0: _DPadOutputValues(0, STICK_PAD_MAX),  # Up
		1: _DPadOutputValues(STICK_PAD_MAX, STICK_PAD_MAX),  # UpRight
		2: _DPadOutputValues(STICK_PAD_MAX, 0),  # Right
		3: _DPadOutputValues(STICK_PAD_MAX, STICK_PAD_MIN),  # DownRight
		4: _DPadOutputValues(0, STICK_PAD_MIN),  # Down
		5: _DPadOutputValues(STICK_PAD_MIN, STICK_PAD_MIN),  # DownLeft
		6: _DPadOutputValues(STICK_PAD_MIN, 0),  # Left
		7: _DPadOutputValues(STICK_PAD_MIN, STICK_PAD_MAX),  # UpLeft
		8: _DPadOutputValues(0, 0),  # Centered
	}

	DPAD_CENTERED_STATE = _DPadOutputValues(0, 0)
	# Leading byte not included in sent output report data.
	# Needed to add for proper CRC32 computation that controller
	# will accept
	BT_CRC32_HEAD = b"\xa2"
	# Leading byte not included in feature report data.
	# Needed if performing CRC32 computation on returned feature
	# report data (for IMU calibration data)
	BT_CALIBRATION_CRC32_HEAD = b"\xa3"

	flags = (
		ControllerFlags.EUREL_GYROS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.HAS_CPAD
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.NO_GRIPS
	)

	# def __init__(self, device, daemon, handle, config_file, config, test_mode=False):
	def __init__(self, driver, syspath, hidrawdev):
		self.driver = driver
		self.daemon = driver.daemon
		self.syspath = syspath

		super().__init__()
		# super().__init__(device, daemon, handle, config_file, config, test_mode=False)

		self._feedback_output = DualSenseHIDOutputBT(
			operating_mode=OperatingMode.DS5_MODE_BT,
			data_id_byte=0x02,
			physical_effect_control=PhysicalEffectControl.ENABLE_HAPTICS,
			motor_left=0,
			motor_right=0,
		)
		self._feedback_cancel_task = None
		self._outputs = {}
		# Use empty struct for starting state
		self._old_state = DualSenseBTControllerInput()

		self._device_name = hidrawdev.getName()
		self._hidrawdev = hidrawdev
		self._fileno = hidrawdev._device.fileno()
		self._id = self._generate_id() if driver else "-"
		self._previous_quat = [1.0, 0.0, 0.0, 0.0]
		self._delta_time = time.time()
		self._previous_time = time.time()

		# time.sleep(1)
		self._set_operational()
		self.read_serial()
		# self.configure()
		self._poller = self.daemon.get_poller()
		if self._poller:
			self._poller.register(self._fileno, self._poller.POLLIN, self._input)
		self.daemon.get_device_monitor().add_remove_callback(syspath, self.close)
		self.daemon.add_controller(self)

	def get_device_name(self):
		return "DualSense over Bluetooth HIDRaw"

	def get_type(self):
		return "ds5bt_hidraw"

	def apply_config(self, config):
		icon = config["icon"]
		led_level = config["led_level"]
		self.configure(icon=icon, led_level=led_level)

	def configure(self, icon=None, led_level=100):
		# log.debug("CALLED CONFIGURE")
		# return

		lightbar_color = (0.0, 0.0, 1.0)  # blue by default
		if icon:
			basename, ext = icon.rsplit(".", 1)
			parts = basename.rsplit("-", 1)
			if parts:
				raw_idx = parts[-1]
				try:
					icon_idx = int(raw_idx)
				except ValueError:
					pass
				else:
					if icon_idx < len(ICON_COLORS):
						lightbar_color = ICON_COLORS[icon_idx]

		led_level_norm = float(led_level) / 100
		lightbar_color_bytes = tuple(int(color_norm * led_level_norm * 255) for color_norm in lightbar_color)

		# print(lightbar_color_bytes)
		# lightbar_red=lightbar_color_bytes[0],
		# lightbar_green=lightbar_color_bytes[1],
		# lightbar_blue=lightbar_color_bytes[2],
		output = DualSenseHIDOutputBT(
			operating_mode=OperatingMode.DS5_MODE_BT,
			data_id_byte=0x02,
			light_effect_control=LightEffectControl.LIGHTBAR_CONTROL_ENABLE,
			lightbar_red=lightbar_color_bytes[0],
			lightbar_green=lightbar_color_bytes[1],
			lightbar_blue=lightbar_color_bytes[2],
		)

		tempbuffer = bytearray(output)
		# print("ITS A BLUE: {}".format(output.lightbar_blue))
		self._prepare_buffer_crc(tempbuffer)
		self.schedule_output("lightbar", tempbuffer)
		# time.sleep(2)
		self.flush()
		# self._hidrawdev.read(78)
		# feature_data = self._hidrawdev.getFeatureReport(9)
		# time.sleep(2)
		# self._hidrawdev.read(78)

	def _set_operational(self):
		# log.debug("CALLING SET_OPERATIONAL")
		# Get feature report for serial performs initial switch
		# to DS5 mode
		feature_data = self._hidrawdev.getFeatureReport(9)

		init_output = DualSenseHIDOutputBT(
			operating_mode=OperatingMode.DS5_MODE_BT,
			data_id_byte=0x02,
			light_effect_control=0x55,
			lightbar_red=0,
			lightbar_green=0,
			lightbar_blue=200,
		)

		tempman = bytearray(init_output)
		"""calcCrc32 = zlib.crc32(b"\xa2") & 0xFFFFFFFF
		calcCrc32 = zlib.crc32(tempman[0:74], calcCrc32) & 0xFFFFFFFF
		tempman[74] = calcCrc32 & 0xFF
		tempman[75] = (calcCrc32 >> 8) & 0xFF
		tempman[76] = (calcCrc32 >> 16) & 0xFF
		tempman[77] = (calcCrc32 >> 24) & 0xFF
		"""
		# self._hidrawdev.write(tempman)
		# time.sleep(1)
		self._prepare_buffer_crc(tempman)
		self.schedule_output("init", tempman)
		self.flush()
		# Seems to not register until the next device read. Need
		# to see if there is a way around that
		self._hidrawdev.read(78)
		# feature_data = self._hidrawdev.getFeatureReport(9)
		# time.sleep(2)

	def _prepare_buffer_crc(self, buf):
		calcCrc32 = zlib.crc32(b"\xa2") & 0xFFFFFFFF
		calcCrc32 = zlib.crc32(buf[0:74], calcCrc32) & 0xFFFFFFFF
		buf[74] = calcCrc32 & 0xFF
		buf[75] = (calcCrc32 >> 8) & 0xFF
		buf[76] = (calcCrc32 >> 16) & 0xFF
		buf[77] = (calcCrc32 >> 24) & 0xFF

	def feedback(self, data):
		position, amplitude, period, count = data.data

		normalized_amp = float(amplitude) / 0x8000
		clamped_amp = int(normalized_amp * 0xFF)
		half_amp = int(normalized_amp * 0x80)

		if position == HapticPos.LEFT:
			# NOTE: the left motor is heavier, so we must give it less oomph
			self._feedback_output.motor_left = half_amp
		elif position == HapticPos.RIGHT:
			self._feedback_output.motor_right = clamped_amp
		elif position == HapticPos.BOTH:
			self._feedback_output.motor_right = clamped_amp
			self._feedback_output.motor_left = half_amp

		duration = float(period) * count / 0x10000
		# The motors don't seem to perform reliably when shut off under 50ms
		duration = max(duration, 0.05)

		tempman = bytearray(self._feedback_output)
		self._prepare_buffer_crc(tempman)
		self.schedule_output("feedback", tempman)
		self.flush()

		def clear_feedback(mapper):
			self._feedback_output.motor_right = self._feedback_output.motor_left = 0
			tempman = bytearray(self._feedback_output)
			self._prepare_buffer_crc(tempman)
			self.schedule_output("feedback", tempman)
			self.flush()

		if self._feedback_cancel_task:
			self._feedback_cancel_task.cancel()
		self._feedback_cancel_task = self.mapper.schedule(duration, clear_feedback)

	def _input(self, *a):
		# log.debug("FOUND INPUT")
		tempdata = self._hidrawdev.read(78)
		# Skip over packet if not a DS5 mode input packet
		if tempdata[0] != 0x31:
			return

		# log.debug(tempdata)
		old_state = self._old_state
		current_time = time.time()
		self._delta_time = current_time - self._previous_time
		# hamtaro = DualSenseHIDInputBT.from_buffer_copy(tempdata)
		# log.debug("LX: {} {}".format(hamtaro.lx, tempdata[2]))
		state_data = self._convert_input_data(tempdata)
		"""print("INPUT [", end="")
		for index, value in enumerate(tempdata):
			print("[{},{}]".format(index, value), end=", ")

		print("")
		"""

		# log.debug(self._hidrawdev._device)
		# self.flush()
		# self.close()
		if self.mapper:
			self.mapper.input(self, old_state, state_data)

		self._old_state = state_data
		# Check for pending output data
		self.flush()

		self._previous_time = current_time

	def _convert_input_data(self, data):
		state = DualSenseBTControllerInput()
		state.stick_x = self._stick_axis_scale(data[2], False)
		state.stick_y = self._stick_axis_scale(data[3], True)
		state.rpad_x = self._stick_axis_scale(data[4], False)
		state.rpad_y = self._stick_axis_scale(data[5], True)
		state.ltrig = data[6]
		state.rtrig = data[7]
		tempbyte = data[9]
		if (tempbyte & (1 << 7)) != 0:
			state.buttons |= SCButtons.Y
		if (tempbyte & (1 << 6)) != 0:
			state.buttons |= SCButtons.B
		if (tempbyte & (1 << 5)) != 0:
			state.buttons |= SCButtons.A
		if (tempbyte & (1 << 4)) != 0:
			state.buttons |= SCButtons.X
		dpad_state = data[9] & 0x0F
		if dpad_state != 8:
			state.buttons |= SCButtons.LPAD | SCButtons.LPADTOUCH
			tempDPad = DS5HidRawController.DPAD_STATE_TYPES.get(dpad_state, DS5HidRawController.DPAD_CENTERED_STATE)
			state.lpad_x = tempDPad.x
			state.lpad_y = tempDPad.y

		tempbyte = data[10]
		if (tempbyte & (1 << 7)) != 0:
			state.buttons |= SCButtons.RPAD
		if (tempbyte & (1 << 6)) != 0:
			state.buttons |= SCButtons.STICKPRESS
		if (tempbyte & (1 << 5)) != 0:
			state.buttons |= SCButtons.START
		if (tempbyte & (1 << 4)) != 0:
			state.buttons |= SCButtons.BACK
		if (tempbyte & (1 << 3)) != 0:
			state.buttons |= SCButtons.RT
		if (tempbyte & (1 << 2)) != 0:
			state.buttons |= SCButtons.LT
		if (tempbyte & (1 << 1)) != 0:
			state.buttons |= SCButtons.RB
		if (tempbyte & (1 << 0)) != 0:
			state.buttons |= SCButtons.LB

		tempbyte = data[11]
		if (tempbyte & (1 << 0)) != 0:
			state.buttons |= SCButtons.C
		if (tempbyte & (1 << 1)) != 0:
			state.buttons |= SCButtons.CPADPRESS

		# Change gyro dir values to match Steam Controller
		state.gpitch = ctypes.c_int16((data[18] << 8) | data[17]).value
		state.gyaw = ctypes.c_int16((data[20] << 8) | data[19]).value * -1
		state.groll = ctypes.c_int16((data[22] << 8) | data[21]).value * -1

		# Change accel axes to match Steam Controller (flip pitch and roll)
		# Scale values for 2G instead of 1G
		state.accel_x = ctypes.c_int16((data[24] << 8) | data[23]).value * 2
		# Invert pitch
		state.accel_y = ctypes.c_int16((data[28] << 8) | data[27]).value * -2
		state.accel_z = ctypes.c_int16((data[26] << 8) | data[25]).value * 2
		# print("GYRO: {} {} {}".format(state.gyaw, state.gpitch, state.groll))
		# print("ACCEL: {} | {} | {}".format(state.accel_x, state.accel_y, state.accel_z))
		# Calculate quaternion for gyro data. Needed for tilt controls output.
		# TODO: Try to add sensor fusion and complementary filter later
		self._calculate_quaternion(state)

		# Check for CPAD touch
		if (data[34] & 0x80) == 0:
			state.buttons |= SCButtons.CPADTOUCH

		state.cpad_x = ((data[36] & 0x0F) << 8) | data[35]
		state.cpad_y = ((data[37] & 0x0F) << 4) | ((data[36] & 0xF0) >> 4)

		return state

	def _stick_axis_scale(self, value, invert=False, test=False):
		result = value - 128
		tempRatio = (result / 127.0) if (value >= 128) else (result / (128.0))
		if invert:
			tempRatio = -tempRatio

		tempRatio = (tempRatio + 1.0) * 0.5

		"""if test:
			print(tempRatio)
			print("RES {}".format(STICK_PAD_RES))
			print(tempRatio * STICK_PAD_RES + STICK_PAD_MIN)
		"""
		result = int(tempRatio * STICK_PAD_RES + STICK_PAD_MIN)
		return result

	def _calculate_quaternion(self, state):
		# Convert raw gyro values to degrees per second
		GYRO_RES_IN_DEG_SEC = 16
		(yaw, pitch, roll) = (
			(state.gyaw / GYRO_RES_IN_DEG_SEC),
			(state.gpitch / GYRO_RES_IN_DEG_SEC),
			(state.groll / GYRO_RES_IN_DEG_SEC),
		)

		# Remove time delta element to get gyro angles and convert to radians
		old_yaw = yaw
		yaw = yaw * self._delta_time
		yaw_rad = yaw * math.pi / 180.0
		pitch = pitch * self._delta_time
		pitch_rad = pitch * math.pi / 180.0
		roll = roll * self._delta_time
		roll_rad = roll * math.pi / 180.0
		# print("GYRO: {} {} {}".format(yaw, pitch, roll))
		# print("GYRO: {} | {} | {} | {} | {}".format(state.gyaw, old_yaw, yaw, yaw_rad, self._delta_time))

		# Obtain current quaternion
		# qx (Roll), qy (Pitch), qz (Yaw), qw (Theta)
		qx = math.sin(roll_rad / 2) * math.cos(pitch_rad / 2) * math.cos(yaw_rad / 2) - math.cos(
			roll_rad / 2,
		) * math.sin(pitch_rad / 2) * math.sin(yaw_rad / 2)
		qy = math.cos(roll_rad / 2) * math.sin(pitch_rad / 2) * math.cos(yaw_rad / 2) + math.sin(
			roll_rad / 2,
		) * math.cos(pitch_rad / 2) * math.sin(yaw_rad / 2)
		qz = math.cos(roll_rad / 2) * math.cos(pitch_rad / 2) * math.sin(yaw_rad / 2) - math.sin(
			roll_rad / 2,
		) * math.sin(pitch_rad / 2) * math.cos(yaw_rad / 2)
		qw = math.cos(roll_rad / 2) * math.cos(pitch_rad / 2) * math.cos(yaw_rad / 2) + math.sin(
			roll_rad / 2,
		) * math.sin(pitch_rad / 2) * math.sin(yaw_rad / 2)

		# Multiply previous calculated quaternion by new quaternion
		(old_qw, old_qx, old_qy, old_qz) = self._previous_quat
		Q0Q1_w = old_qw * qw - old_qx * qx - old_qy * qy - old_qz * qz
		Q0Q1_x = old_qw * qx + old_qx * qw + old_qy * qz - old_qz * qy
		Q0Q1_y = old_qw * qy - old_qx * qz + old_qy * qw + old_qz * qx
		Q0Q1_z = old_qw * qz + old_qx * qy - old_qy * qx + old_qz * qw

		# Convert normalized values to mapper expected range and store
		# in state object
		# q1 (Theta), q2 (Pitch), q3 (Roll), q4 (Yaw)
		state.q1 = int(Q0Q1_w * 32767.0)
		state.q2 = int(Q0Q1_y * 32767.0)
		state.q3 = int(Q0Q1_x * 32767.0)
		# Invert Yaw to match Steam Controller
		state.q4 = int(Q0Q1_z * -32767.0)
		# Store calculated quaternion for next poll
		self._previous_quat = [Q0Q1_w, Q0Q1_x, Q0Q1_y, Q0Q1_z]
		# print("TEST QUAT: {}".format(self._previous_quat))
		# print("TEST QUAT Z: {}".format(self._previous_quat[3] * -1))

	def close(self):
		if self._poller:
			self._poller.unregister(self._fileno)

		self.daemon.remove_controller(self)
		self._hidrawdev._device.close()
		# log.debug("CLOSING")

	def read_serial(self):
		self._serial = self._hidrawdev.getPhysicalAddress().replace(b":", b"")

	def schedule_output(self, output_id, output):
		self._outputs[output_id] = output

	def flush(self):
		while self._outputs:
			output_id, output = self._outputs.popitem()
			# print("PAYLOAD {} {}".format(output_id, output))
			self._hidrawdev.write(output)
			# time.sleep(0.1)
			# print("")

		# print("SLEEPING")
		# time.sleep(0.5)

	# TODO: Remove. Temporarily keep as a reference to the nonsense
	# I went through to figure out how to compute valid CRC32 that
	# the controller would accept
	def flushni(self):
		output = []
		print(hex(zlib.crc32(b"hello-world") & 0xFFFFFFFF))

		feedback_output2 = DualSenseHIDOutputBT(
			operating_mode=0x31,
			data_id_byte=0x02,
			physical_effect_control=0x0F,
			light_effect_control=0x55,
			motor_right=0x00,
			brightlite=0x06,
			lightbar_control=0x02,
			led_brightness=0x02,
			lightbar_red=200,
			lightbar_green=90,
			lightbar_blue=211,
		)

		calcCrc32 = zlib.crc32(b"\xa2") & 0xFFFFFFFF
		# calcCrc32 = ~zlib.crc32(b"\x31", calcCrc32)
		log.debug("CRC BETA")
		log.debug(calcCrc32)

		log.debug(len(bytearray(feedback_output2)[0:74]))
		damn = [int(i) for i in bytearray(feedback_output2)[0:78]]
		log.debug(damn)
		# Working
		# tempman = bytearray([0x31, 0x02, 0x0f, 0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x02, 0x02, 0x00, 0x14, 0x14, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
		tempman = bytearray(feedback_output2)
		# Working
		# calcCrc32 = binascii.crc32(bytearray([0x31, 0x02, 0x0f, 0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x02, 0x02, 0x00, 0x14, 0xFF, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00]), ~calcCrc32)
		# calcCrc32 = binascii.crc32(tempman[0:74], ~calcCrc32)
		calcCrc32 = zlib.crc32(tempman[0:74], calcCrc32) & 0xFFFFFFFF
		# _feedback_output.crc32 = calcCrc32
		# _feedback_output.crc32[0] = calcCrc32
		# _feedback_output.crc32[1] = calcCrc32 >> 8
		# _feedback_output.crc32[2] = calcCrc32 >> 16
		# _feedback_output.crc32[3] = calcCrc32 >> 24
		tempman[74] = calcCrc32 & 0xFF
		tempman[75] = (calcCrc32 >> 8) & 0xFF
		tempman[76] = (calcCrc32 >> 16) & 0xFF
		tempman[77] = (calcCrc32 >> 24) & 0xFF
		log.debug("CRC")
		log.debug(calcCrc32)

		print("RADIO EDIT")
		# print(bytearray(_feedback_output))
		"""print("[", end="")
		for index, value in enumerate(tempman):
			print("[{},{}]".format(index, value), end=", ")

		print("")
		"""

		# _test = bytes(_feedback_output)
		# _test = bytearray([0x31, 0x02, 0x0f, 0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x02, 0x02, 0x00, 0x14, 0x14, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])
		# _test[74] = calcCrc32 & 0xFF
		# _test[75] = (calcCrc32 >> 8) & 0xFF
		# _test[76] = (calcCrc32 >> 16) & 0xFF
		# _test[77] = (calcCrc32 >> 24) & 0xFF
		# print(len(_test))
		# print("THE THING {} {} {}".format(_test[40], _test[46], _test[77]))

		result = self._hidrawdev.write(tempman)
		time.sleep(0.1)
		print(result)

	def get_type(self):
		return "ds5"

	def get_gui_config_file(self):
		return "ds5-config.json"

	def _generate_id(self):
		"""ID is generated as 'ds5' or 'ds5:X' where 'X' starts as 1 and increases
		as controllers with same ids are connected.
		"""
		magic_number = 1
		id = "ds5"
		while id in self.daemon.get_active_ids():
			id = "ds5:%s" % (magic_number,)
			magic_number += 1
		return id

	def get_gyro_enabled(self):
		# Cannot be actually turned off, so it's always active
		# TODO: Maybe emulate turning off?
		return True


class DS5EvdevController(EvdevController):
	TOUCH_FACTOR_X = STICK_PAD_MAX / 940.0
	TOUCH_FACTOR_Y = STICK_PAD_MAX / 470.0
	BUTTON_MAP = {
		304: "A",
		305: "B",
		307: "Y",
		308: "X",
		310: "LB",
		311: "RB",
		# TODO: Figure out what it is the purpose of the button event when using the trigger
		# 312: "LT2",
		# 313: "RT2",
		314: "BACK",
		315: "START",
		316: "C",
		317: "STICKPRESS",
		318: "RPAD",
		# 319: "CPAD",
	}
	AXIS_MAP = {
		0: {"axis": "stick_x", "deadzone": 4, "max": 255, "min": 0},
		1: {"axis": "stick_y", "deadzone": 4, "max": 0, "min": 255},
		3: {"axis": "rpad_x", "deadzone": 4, "max": 255, "min": 0},
		4: {"axis": "rpad_y", "deadzone": 8, "max": 0, "min": 255},
		2: {"axis": "ltrig", "max": 255, "min": 0},
		5: {"axis": "rtrig", "max": 255, "min": 0},
		16: {"axis": "lpad_x", "deadzone": 0, "max": 1, "min": -1},
		17: {"axis": "lpad_y", "deadzone": 0, "max": -1, "min": 1},
	}
	# TODO: Should the old button for DS4 map be removed? DualSense support came with kernel 5.12
	# BUTTON_MAP_OLD = {
	# 	304: "X",
	# 	305: "A",
	# 	306: "B",
	# 	307: "Y",
	# 	308: "LB",
	# 	309: "RB",
	# 	312: "BACK",
	# 	313: "START",
	# 	314: "STICKPRESS",
	# 	315: "RPAD",
	# 	316: "C",
	# 	# 317: "CPAD",
	# }
	# AXIS_MAP_OLD = {
	# 	0:  { "axis": "stick_x", "deadzone": 4, "max": 255, "min": 0 },
	# 	1:  { "axis": "stick_y", "deadzone": 4, "max": 0, "min": 255 },
	# 	2:  { "axis": "rpad_x", "deadzone": 4, "max": 255, "min": 0 },
	# 	5:  { "axis": "rpad_y", "deadzone": 8, "max": 0, "min": 255 },
	# 	3:  { "axis": "ltrig", "max": 32767, "min": -32767 },
	# 	4:  { "axis": "rtrig", "max": 32767, "min": -32767 },
	# 	16: { "axis": "lpad_x", "deadzone": 0, "max": 1, "min": -1 },
	# 	17: { "axis": "lpad_y", "deadzone": 0, "max": -1, "min": 1 }
	# }
	GYRO_MAP = {
		EvdevController.ECODES.ABS_RX: ("gpitch", 0.01),
		EvdevController.ECODES.ABS_RY: ("gyaw", 0.01),
		EvdevController.ECODES.ABS_RZ: ("groll", 0.01),
		EvdevController.ECODES.ABS_X: (None, 1),  # 'q2'
		EvdevController.ECODES.ABS_Y: (None, 1),  # 'q3'
		EvdevController.ECODES.ABS_Z: (None, -1),  # 'q1'
	}
	flags = (
		ControllerFlags.EUREL_GYROS
		| ControllerFlags.HAS_RSTICK
		| ControllerFlags.HAS_CPAD
		| ControllerFlags.HAS_DPAD
		| ControllerFlags.SEPARATE_STICK
		| ControllerFlags.NO_GRIPS
	)

	def __init__(self, daemon, controllerdevice, gyro, touchpad):
		config = {"axes": DS5EvdevController.AXIS_MAP, "buttons": DS5EvdevController.BUTTON_MAP, "dpads": {}}
		# if controllerdevice.info.version & 0x8000 == 0:
		# 	# Older kernel uses different mappings
		# 	# see kernel source, drivers/hid/hid-sony.c#L2748
		# 	config['axes'] = DS4EvdevController.AXIS_MAP_OLD
		# 	config['buttons'] = DS4EvdevController.BUTTON_MAP_OLD
		self._gyro = gyro
		self._touchpad = touchpad
		for device in (self._gyro, self._touchpad):
			if device:
				device.grab()
		EvdevController.__init__(self, daemon, controllerdevice, None, config)
		if self.poller:
			self.poller.register(touchpad.fd, self.poller.POLLIN, self._touchpad_input)
			self.poller.register(gyro.fd, self.poller.POLLIN, self._gyro_input)

	def _gyro_input(self, *a):
		new_state = self._state
		try:
			for event in self._gyro.read():
				if event.type == self.ECODES.EV_ABS:
					axis, factor = DS5EvdevController.GYRO_MAP[event.code]
					if axis:
						new_state = new_state._replace(**{axis: int(event.value * factor)})
		except OSError:
			# Errors here are not even reported, evdev class handles important ones
			return

		if new_state is not self._state:
			old_state, self._state = self._state, new_state
			if self.mapper:
				self.mapper.input(self, old_state, new_state)

	def _touchpad_input(self, *a):
		new_state = self._state
		try:
			for event in self._touchpad.read():
				if event.type == self.ECODES.EV_ABS:
					if event.code == self.ECODES.ABS_MT_POSITION_X:
						value = event.value * DS5EvdevController.TOUCH_FACTOR_X
						value = STICK_PAD_MIN + int(value)
						new_state = new_state._replace(cpad_x=value)
					elif event.code == self.ECODES.ABS_MT_POSITION_Y:
						value = event.value * DS5EvdevController.TOUCH_FACTOR_Y
						value = STICK_PAD_MAX - int(value)
						new_state = new_state._replace(cpad_y=value)
				elif event.type == 0:
					pass
				elif event.code == self.ECODES.BTN_LEFT:
					if event.value == 1:
						b = new_state.buttons | SCButtons.CPADPRESS
						new_state = new_state._replace(buttons=b)
					else:
						b = new_state.buttons & ~SCButtons.CPADPRESS
						new_state = new_state._replace(buttons=b)
				elif event.code == self.ECODES.BTN_TOUCH:
					if event.value == 1:
						b = new_state.buttons | SCButtons.CPADTOUCH
						new_state = new_state._replace(buttons=b)
					else:
						b = new_state.buttons & ~SCButtons.CPADTOUCH
						new_state = new_state._replace(buttons=b, cpad_x=0, cpad_y=0)
		except OSError:
			# Errors here are not even reported, evdev class handles important ones
			return

		if new_state is not self._state:
			old_state, self._state = self._state, new_state
			if self.mapper:
				self.mapper.input(self, old_state, new_state)

	def close(self):
		EvdevController.close(self)
		for device in (self._gyro, self._touchpad):
			try:
				self.poller.unregister(device.fd)
				device.ungrab()
			except:
				pass

	def get_gyro_enabled(self):
		# Cannot be actually turned off, so it's always active
		# TODO: Maybe emulate turning off?
		return True

	def get_type(self):
		return "ds5evdev"

	# TODO: Create ds5-config.json for GUI
	def get_gui_config_file(self):
		return "ds5-config.json"

	def __repr__(self):
		return "<DS5EvdevController %s>" % (self.get_id(),)

	def _generate_id(self):
		"""ID is generated as 'ds5' or 'ds5:X' where 'X' starts as 1 and increases as controllers with same ids are connected."""
		magic_number = 1
		id = "ds5"
		while id in self.daemon.get_active_ids():
			id = "ds5:%s" % (magic_number,)
			magic_number += 1
		return id


def init(daemon, config):
	"""Register hotplug callback for DS5 device."""

	def hid_callback(device, handle):
		return DS5Controller(device, daemon, handle, None, None)

	def make_evdev_device(syspath: str, *whatever):
		devices = get_evdev_devices_from_syspath(syspath)
		# With kernel 4.10 or later, PS4 controller pretends to be 3 different devices.
		# 1st, determining which one is actual controller is needed
		controllerdevice = None
		for device in devices:
			count = len(get_axes(device))
			if count == 8:
				# 8 axes - Controller
				controllerdevice = device
		if not controllerdevice:
			log.warning("Failed to determine controller device")
			return None
		# 2nd, find motion sensor and touchpad with physical address matching controllerdevice
		gyro, touchpad = None, None
		phys = device.phys.split("/")[0]
		for device in devices:
			if device.phys.startswith(phys):
				axes = get_axes(device)
				count = len(axes)
				if count == 6:
					# 6 axes
					if EvdevController.ECODES.ABS_MT_POSITION_X in axes:
						# kernel 4.17+ - touchpad
						touchpad = device
					else:
						# gyro sensor
						gyro = device
				elif count == 4:
					# 4 axes - Touchpad
					touchpad = device
		# 3rd, do a magic
		if controllerdevice and gyro and touchpad:
			return make_new_device(DS5EvdevController, controllerdevice, gyro, touchpad)

	def fail_cb(syspath: str, vid: int, pid: int):
		if HAVE_EVDEV:
			log.warning("Failed to acquire USB device, falling back to evdev driver. This is far from optimal.")
			make_evdev_device(syspath)
		else:
			log.error(
				"Failed to acquire USB device and evdev is not available. Everything is lost and DS5 support disabled.",
			)
		# TODO: Maybe add_error here, but error reporting needs little rework so it's not threated as fatal
		# daemon.add_error("ds5", "No access to DS5 device")

	if config["drivers"].get("hiddrv") or (HAVE_EVDEV and config["drivers"].get("evdevdrv")):
		for product_id in PRODUCT_IDS:
			register_hotplug_device(hid_callback, VENDOR_ID, product_id, on_failure=fail_cb)
		if config["drivers"].get("hiddrv"):
			# Only enable HIDRaw support for BT connections if hiddrv is enabled
			_drv = DS5HidRawDriver(daemon, config)
		elif HAVE_EVDEV and config["drivers"].get("evdevdrv"):
			# Attempt evdev as a backup
			for product_id in PRODUCT_IDS:
				daemon.get_device_monitor().add_callback(
					"bluetooth",
					VENDOR_ID,
					product_id,
					make_evdev_device,
					None,
				)
		return True
	log.warning("Neither HID nor Evdev driver is enabled, DS5 support cannot be enabled.")
	return False


if __name__ == "__main__":
	""" Called when executed as script """
	init_logging()
	set_logging_level(True, True)
	sys.exit(hiddrv_test(DS5Controller, ["054c:0ce6", "054c:0df2"]))



"""SC Controller - tools

Various stuff that I don't care to fit anywhere else.
"""

from __future__ import annotations

import ctypes
import importlib.machinery
import logging
import os
import shlex
import sysconfig
from math import atan2, cos, sin, sqrt
from math import pi as PI
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from collections.abc import Callable
	from typing import Any

from scc.paths import (
	get_button_images_path,
	get_controller_icons_path,
	get_default_controller_icons_path,
	get_default_menuicons_path,
	get_default_menus_path,
	get_default_profiles_path,
	get_menuicons_path,
	get_menus_path,
	get_profiles_path,
)

HAVE_POSIX1E = False
try:
	import posix1e

	HAVE_POSIX1E = True
except ImportError:
	pass

log = logging.getLogger("tools.py")
_ = lambda x: x

LOG_FORMAT = "%(levelname)s %(name)-13s %(message)s"


def init_logging(prefix: str = "", suffix: str = ""):
	"""Initialize logging, set custom logging format and add one logging level with name and method to call.

	prefix and suffix arguments can be used to modify log level prefixes.
	"""
	logging.basicConfig(format=LOG_FORMAT)
	logger = logging.getLogger()
	# Rename levels
	logging.addLevelName(10, prefix + "D" + suffix)  # Debug
	logging.addLevelName(20, prefix + "I" + suffix)  # Info
	logging.addLevelName(30, prefix + "W" + suffix)  # Warning
	logging.addLevelName(40, prefix + "E" + suffix)  # Error
	# Create additional, "verbose" level
	logging.addLevelName(15, prefix + "V" + suffix)  # Verbose

	# Add 'logging.verbose' method
	def verbose(self: logging.Logger, msg: str, *args, **kwargs):
		return self.log(15, msg, *args, **kwargs)

	logging.Logger.verbose = verbose
	# Wrap Logger._log in something that can handle utf-8 exceptions
	old_log = logging.Logger._log

	def _log(self, level, msg, args, exc_info=None, extra=None):
		args = tuple([(c.decode("utf-8") if type(c) is bytes else c) for c in args])
		# msg = msg if type(msg) is str else msg.decode("utf-8")
		old_log(self, level, msg, args, exc_info, extra)

	logging.Logger._log = _log


def set_logging_level(verbose, debug):
	"""Set logging level"""
	logger = logging.getLogger()
	if debug:  # everything
		logger.setLevel(0)
	elif verbose:  # everything but debug
		logger.setLevel(11)
	else:  # INFO and worse
		logger.setLevel(20)


def ensure_size(n: int, lst: list[Any], fill_with: Any = None) -> list[Any]:
	"""Return copy of lst with size 'n'.

	If lst is shorter, None's are appended.
	If lst is longer, it is cut.
	"""
	l = list(lst)
	while len(l) < n:
		l.append(fill_with)
	return l[0:n]


def quat2euler(q0: float, q1: float, q2: float, q3: float) -> tuple[float, float, float]:
	"""Convert quaterion to (pitch, yaw, roll).

	Values are in -PI to PI range.
	"""
	qq0, qq1, qq2, qq3 = q0**2, q1**2, q2**2, q3**2
	xa = qq0 - qq1 - qq2 + qq3
	xb = 2 * (q0 * q1 + q2 * q3)
	xn = 2 * (q0 * q2 - q1 * q3)
	yn = 2 * (q1 * q2 + q0 * q3)
	zn = qq3 + qq2 - qq0 - qq1

	pitch = atan2(xb, xa)
	yaw = atan2(xn, sqrt(1 - xn**2))
	roll = atan2(yn, zn)
	return pitch, yaw, roll


def point_in_gtkrect(rect, x: float, y: float) -> bool:
	return x > rect.x and y > rect.y and x < rect.x + rect.width and y < rect.y + rect.height


def anglediff(a1: float, a2: float) -> float:
	"""Expects values in radians"""
	return (a2 - a1 + PI) % (2.0 * PI) - PI


def degdiff(a1: float, a2: float) -> float:
	"""Expects values in degrees"""
	return (a2 - a1 + 180.0) % 360.0 - 180.0


def nameof(e: Any) -> str:
	"""If 'e' is enum value, return e.name.

	Otherwise, return str(e).
	"""
	return e.name if hasattr(e, "name") else str(e)


def shjoin(lst: list) -> bytes:
	"""Joins list into shell-escaped, utf-8 encoded string"""
	s = [x.encode("utf-8") for x in lst]
	#   - escape quotes
	s = [x.encode("unicode_escape") if (b'"' in x or b"'" in x) else x for x in s]
	#   - quote strings with spaces
	s = [b"'%s'" % (x,) if b" " in x else x for x in s]
	return b" ".join(s)


def shsplit(s: str) -> list[str]:
	"""Returns original list from what shjoin returned"""
	lex = shlex.shlex(s, posix=True)
	lex.escapedquotes = "\"'"
	lex.whitespace_split = True
	return [x for x in list(lex)]


def static_vars(**kwargs):
	"""Static variable func decorator"""

	def decorate(func):
		"""Inner function used to add kwargs attribute to a func"""
		for k in kwargs:
			setattr(func, k, kwargs[k])
		return func

	return decorate


def profile_is_override(name: str) -> bool:
	"""Returns True if named profile exists both in user config directory and
	default_profiles directory.
	"""
	filename = "%s.sccprofile" % (name,)
	if os.path.exists(os.path.join(get_profiles_path(), filename)):
		if os.path.exists(os.path.join(get_default_profiles_path(), filename)):
			return True
	return False


def profile_is_default(name: str) -> bool:
	"""Returns True if named profile exists in default_profiles directory, even
	if it is overrided by profile in user config directory.
	"""
	filename = "%s.sccprofile" % (name,)
	return os.path.exists(os.path.join(get_default_profiles_path(), filename))


def get_profile_name(path: str) -> str:
	"""Returns profile name for specified path. Basically removes path and
	.sccprofile and .mod extension.
	"""
	parts = os.path.split(path)[-1].split(".")
	if parts[-1] == "mod":
		parts = parts[0:-1]
	if parts[-1] == "sccprofile":
		parts = parts[0:-1]
	return ".".join(parts)


def find_profile(name: str) -> str | None:
	"""Returns filename for specified profile name.
	This is done by searching for name + '.sccprofile' in ~/.config/scc/profiles
	first and in /usr/share/scc/default_profiles if file is not found in first
	location.

	Returns None if profile cannot be found.
	"""
	filename = "%s.sccprofile" % (name,)
	for p in (get_profiles_path(), get_default_profiles_path()):
		path = os.path.join(p, filename)
		if os.path.exists(path):
			return path
	return None


def find_icon(
	name: str | None,
	prefer_bw: bool = False,
	paths: tuple[str, ...] | None = None,
	extensions: tuple[str, ...] = ("png", "svg"),
) -> tuple[None, bool] | tuple[str, bool]:
	"""Returns (filename, has_colors) for specified icon name.
	This is done by searching for name + '.png' and name + ".bw.png"
	in user and default menu-icons folders. ".svg" is also supported, but only
	if no pngs are found.

	If both colored and grayscale version is found, colored is returned, unless
	prefer_bw is set to True.

	paths defaults to icons for menuicons

	Returns (None, False) if icon cannot be found.
	"""
	if name is None:
		# Special case, so code can pass menuitem.icon directly
		return None, False
	if paths is None:
		paths = get_default_menuicons_path(), get_menuicons_path()
	if name.endswith(".bw"):
		name = name[0:-3]
	for extension in extensions:
		gray_filename = "%s.bw.%s" % (name, extension)
		colors_filename = "%s.%s" % (name, extension)
		gray, colors = None, None
		for p in paths:
			# Check grayscale
			if gray is None:
				path = os.path.join(p, gray_filename)
				if os.path.exists(path):
					if prefer_bw:
						return path, False
					gray = path
			# Check colors
			if colors is None:
				path = os.path.join(p, colors_filename)
				if os.path.exists(path):
					if not prefer_bw:
						return path, True
					colors = path
		if colors is not None:
			return colors, True
		if gray is not None:
			return gray, False
	return None, False


def find_button_image(
	name: str | None, prefer_bw: bool = False,
) -> tuple[None, bool] | tuple[str, bool]:
	"""Similar to find_icon, but searches for button image"""
	return find_icon(nameof(name), prefer_bw, paths=[get_button_images_path()], extensions=("svg",))


def menu_is_default(name: str) -> bool:
	"""Returns True if named menu exists in default_menus directory, even
	if it is overrided by menu in user config directory.
	"""
	return os.path.exists(os.path.join(get_default_menus_path(), name))


def find_menu(name: str) -> str | None:
	"""Returns filename for specified menu name.
	This is done by searching for name in ~/.config/scc/menus
	first and in /usr/share/scc/default_menus later.

	Returns None if menu cannot be found.
	"""
	for p in (get_menus_path(), get_default_menus_path()):
		path = os.path.join(p, name)
		if os.path.exists(path):
			return path
	return None


def find_controller_icon(name: str) -> str | None:
	"""Returns filename for specified controller icon name.
	This is done by searching for name in ~/.config/controller-icons
	first and in /usr/share/scc/images/controller-icons later.

	Returns None if icon cannot be found.
	"""
	for p in (get_controller_icons_path(), get_default_controller_icons_path()):
		path = os.path.join(p, name)
		if os.path.exists(path):
			return path
	return None


def find_binary(name: str) -> str:
	"""Returns full path to script or binary.

	With some exceptions, this is done simply by searching PATH environment variable.
	"""
	if name.startswith("scc-osd-daemon"):
		# Special case, this one is not supposed to go to /usr/bin
		return os.path.join(os.path.split(__file__)[0], "x11", "scc-osd-daemon.py")
	if name.startswith("scc-autoswitch-daemon"):
		# As above
		return os.path.join(os.path.split(__file__)[0], "x11", "scc-autoswitch-daemon.py")
	user_path = os.environ["PATH"].split(":")
	# Try to add the standard binary paths if not present in PATH
	for d in ["/sbin", "/bin", "/usr/sbin", "/usr/bin"]:
		if d not in user_path:
			user_path.append(d)
	for i in user_path:
		path = os.path.join(i, name)
		if os.path.exists(path):
			return path
	# Not found, return name back and hope for miracle
	return name


def find_library(libname: str) -> ctypes.CDLL:
	"""Search for 'libname.so'.

	Return library loaded with ctypes.CDLL
	Raise OSError if library is not found
	"""
	base_path = os.path.dirname(__file__)
	lib, search_paths = None, []
	so_extensions = importlib.machinery.EXTENSION_SUFFIXES
	site_packages_path = sysconfig.get_path("purelib")

	# Try looking in site-packages of the current environment, pwd and ../pwd
	for extension in so_extensions:
		search_paths += [
			os.path.abspath(os.path.normpath(os.path.join(base_path, "..", libname + extension))),
			os.path.abspath(os.path.normpath(os.path.join(base_path, "../..", libname + extension))),
			os.path.abspath(os.path.normpath(os.path.join(site_packages_path, libname + extension))),
		]

	for path in search_paths:
		if os.path.exists(path):
			lib = path
			break

	if not lib:
		raise OSError("Cant find {}.so. searched at:\n  {}".format(libname, "\n  ".join(search_paths)))
	return ctypes.CDLL(lib)


def find_gksudo() -> list[str] | None:
	"""Search for gksudo or other known graphical sudoing tool.

	Return a list of arguments.
	"""
	SUDOS = ["gksudo", "gksu", "kdesudo", "pkexec", "xdg-su"]
	for name in SUDOS:
		args = name.split(" ")
		bin = find_binary(args[0])
		if bin != args[0]:
			return args
	return None


def check_access(filename: str, write_required: bool = True) -> bool:
	"""Check if user has read and optionally write access to the specified file.

	Use acl first and posix file permisions if acl cannot be used.
	Return true only if user has both required access rights.
	"""
	if HAVE_POSIX1E:
		for pset in posix1e.ACL(file=filename):
			if pset.tag_type == posix1e.ACL_USER and pset.qualifier == os.geteuid():
				if pset.permset.test(posix1e.ACL_READ) and (not write_required or pset.permset.test(posix1e.ACL_WRITE)):
					return True
			if pset.tag_type == posix1e.ACL_GROUP and pset.qualifier in os.getgroups():
				if pset.permset.test(posix1e.ACL_READ) and (not write_required or pset.permset.test(posix1e.ACL_WRITE)):
					return True
	if write_required:
		return os.access(filename, os.R_OK | os.W_OK)
	return os.access(filename, os.R_OK)


def strip_gesture(gstr: str):
	"""Converts gesture string to version where stroke lenght is ignored.

	That means removing repeating characters and adding 'i' to front.
	"""
	last: None | str = None
	uniq: list[str] = []
	for x in gstr:
		if x != last:
			uniq.append(x)
		last = x
	if uniq[0] != "i":
		uniq = ["i"] + uniq
	return "".join(uniq)


clamp: Callable[[float, float, float], float] = lambda low, value, high: min(high, max(low, value))

PId4 = PI / 4.0


def circle_to_square(x: float, y: float) -> tuple[float, float]:
	"""Projects coordinate in circle (of radius 1.0) to coordinate in square.
	"""
	# Adapted from http://theinstructionlimit.com/squaring-the-thumbsticks

	# Determine the theta angle
	angle = atan2(y, x) + PI

	squared = 0, 0
	# Scale according to which wall we're clamping to
	# X+ wall
	if angle <= PId4 or angle > 7.0 * PId4:
		squared = x * (1.0 / cos(angle)), y * (1.0 / cos(angle))
	# Y+ wall
	elif angle > PId4 and angle <= 3.0 * PId4:
		squared = x * (1.0 / sin(angle)), y * (1.0 / sin(angle))
	# X- wall
	elif angle > 3.0 * PId4 and angle <= 5.0 * PId4:
		squared = x * (-1.0 / cos(angle)), y * (-1.0 / cos(angle))
	# Y- wall
	elif angle > 5.0 * PId4 and angle <= 7.0 * PId4:
		squared = x * (-1.0 / sin(angle)), y * (-1.0 / sin(angle))
	else:
		raise ValueError("Invalid angle...?")

	return squared



"""SC-Controller - Aliases.

This module generates Keys.BTN_x and Axes.AXIS_x aliases when imported
"""

from scc.uinput import Axes, Keys

ALL_BUTTONS = (
	Keys.BTN_START,
	Keys.BTN_MODE,
	Keys.BTN_SELECT,
	Keys.BTN_A,
	Keys.BTN_B,
	Keys.BTN_X,
	Keys.BTN_Y,
	Keys.BTN_TL,
	Keys.BTN_TR,
	Keys.BTN_THUMBL,
	Keys.BTN_THUMBR,
	Keys.BTN_WHEEL,
	Keys.BTN_GEAR_DOWN,
	Keys.BTN_GEAR_UP,
	Keys.KEY_OK,
	Keys.KEY_SELECT,
	Keys.KEY_GOTO,
	Keys.KEY_CLEAR,
	Keys.KEY_OPTION,
	Keys.KEY_INFO,
	Keys.KEY_TIME,
	Keys.KEY_VENDOR,
	Keys.KEY_ARCHIVE,
	Keys.KEY_PROGRAM,
	Keys.KEY_CHANNEL,
	Keys.KEY_FAVORITES,
	Keys.KEY_EPG,
)

ALL_AXES = (
	Axes.ABS_X,
	Axes.ABS_Y,
	Axes.ABS_RX,
	Axes.ABS_RY,
	Axes.ABS_Z,
	Axes.ABS_RZ,
	Axes.ABS_HAT0X,
	Axes.ABS_HAT0Y,
	Axes.ABS_HAT1X,
	Axes.ABS_HAT1Y,
	Axes.ABS_HAT2X,
	Axes.ABS_HAT2Y,
	Axes.ABS_HAT3X,
	Axes.ABS_HAT3Y,
	Axes.ABS_PRESSURE,
	Axes.ABS_DISTANCE,
	Axes.ABS_TILT_X,
	Axes.ABS_TILT_Y,
	Axes.ABS_TOOL_WIDTH,
	Axes.ABS_VOLUME,
	Axes.ABS_MISC,
)

for i in range(len(ALL_BUTTONS)):
	setattr(Keys, "BTN%i" % (i,), ALL_BUTTONS[i])

for i in range(len(ALL_AXES)):
	setattr(Axes, "ABS%i" % (i,), ALL_AXES[i])



#!/usr/bin/env python3
"""Script that integrates everything SCC can do in one executable.

Created so scc-* stuff doesn't polute /usr/bin.
"""

from scc.scripts import main

if __name__ == "__main__":
	main()



#!/usr/bin/env python3
"""SC-Controller - OSD Message

Display message that just sits there
"""


import logging

from gi.repository import GLib, Gtk

from scc.osd import OSDWindow
from scc.special_actions import OSDAction

log = logging.getLogger("osd.message")


class Message(OSDWindow):
	def __init__(self):
		OSDWindow.__init__(self, "osd-message")

		self.timeout = OSDAction.DEFAULT_TIMEOUT
		self.size = OSDAction.DEFAULT_SIZE
		self.text = "text"
		self._timeout_id = None

	def show(self):
		self.l = Gtk.Label()
		self.l.set_name("osd-label-%s" % (self.size,))
		self.l.set_label(self.text)

		self.add(self.l)

		if self.size < 2:
			self.set_name("osd-message-1")
		OSDWindow.show(self)
		if self.timeout > 0:
			self._timeout_id = GLib.timeout_add_seconds(self.timeout, self.quit)

	def extend(self):
		self.set_state(Gtk.StateType.ACTIVE)
		self.l.set_state(Gtk.StateType.ACTIVE)
		GLib.timeout_add_seconds(0.5, self.cancel_active_state)
		if self._timeout_id:
			GLib.source_remove(self._timeout_id)
			self._timeout_id = GLib.timeout_add_seconds(self.timeout, self.quit)

	def cancel_active_state(self):
		self.set_state(Gtk.StateType.NORMAL)
		self.l.set_state(Gtk.StateType.NORMAL)

	def hash(self):
		return hash(self.text) + self.timeout - (self.size * 5)

	def _add_arguments(self):
		OSDWindow._add_arguments(self)
		self.argparser.add_argument(
			"-t",
			type=float,
			metavar="seconds",
			default=5,
			help="time before message is hidden (default: 5; 0 means forever)",
		)
		self.argparser.add_argument(
			"-s", type=int, metavar="size", default=3, help="font size, in range 1 to 3 (default: 3)",
		)
		self.argparser.add_argument("text", type=str, help="text to display")

	def parse_argumets(self, argv):
		if not OSDWindow.parse_argumets(self, argv):
			return False
		self.text = self.args.text
		self.timeout = self.args.t
		self.size = self.args.s
		return True



"""SC-Controller - Scripts.

Contains code for most of what can be done using 'scc' script.
Created so scc-* stuff doesn't polute /usr/bin.
"""

from __future__ import annotations

import os
import subprocess
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from io import TextIOWrapper


from scc.tools import find_binary, init_logging, set_logging_level


class InvalidArguments(Exception):
	pass


def run_binary(binary_name: str, argv: list[str]) -> int:
	"""Run scc-daemon with passed parameters."""
	binary = find_binary(binary_name)
	child = subprocess.Popen([binary] + argv)
	child.communicate()
	return child.returncode


def cmd_daemon(argv0: str, argv: list[str]) -> int:
	"""Run scc-daemon with passed parameters."""
	return run_binary("scc-daemon", argv)


def help_daemon() -> int:
	"""Run scc-daemon --help."""
	return run_binary("scc-daemon", ["--help"])


def cmd_gui(argv0: str, argv: list[str]) -> int:
	"""Run sc-controller(GUI) with passed parameters."""
	return run_binary("sc-controller", argv)


def help_gui() -> int:
	"""Run sc-controller --help."""
	return run_binary("sc-controller", ["--help"])


def cmd_test_evdev(argv0: str, argv: list[str]) -> int:
	"""Evdev driver test. Displays gamepad inputs using evdev driver.

	Usage: scc test-evdev /dev/input/node
	Return codes:
	  0 - normal exit
	  1 - invalid arguments or other error
	  2 - failed to open device
	"""
	from scc.drivers.evdevdrv import evdevdrv_test

	return evdevdrv_test(argv)


def cmd_test_hid(argv0: str, argv: list[str]) -> int:
	"""HID driver test. Displays gamepad inputs using hid driver.

	Usage: scc test-hid vendor_id device_id
	Return codes:
	  0 - normal exit
	  1 - invalid arguments or other error
	  2 - failed to open device
	  3 - device is not HID-compatibile
	  4 - failed to parse HID descriptor
	"""
	from scc.drivers.hiddrv import HIDController, hiddrv_test

	return hiddrv_test(HIDController, argv)


def help_osd_keyboard() -> None:
	import_osd()
	from scc.osd.keyboard import Keyboard

	return run_osd_tool(Keyboard(), "osd-keyboard", ["--help"])


def cmd_osd_keyboard(argv0: str, argv: list[str]) -> None:
	"""Display on-screen keyboard."""
	import_osd()
	from scc.osd.keyboard import Keyboard

	return run_osd_tool(Keyboard(), argv0, argv)


def cmd_list_profiles(argv0: str, argv: list[str]) -> int:
	"""List available profiles.

	Usage: scc list-profiles [-a]

	Arguments:
		-a   Include names begining with dot

	"""
	from scc.paths import get_default_profiles_path, get_profiles_path

	paths = [get_default_profiles_path(), get_profiles_path()]
	include_hidden = "-a" in argv
	lst = set()
	for path in paths:
		try:
			for x in os.listdir(path):
				if x.endswith(".sccprofile"):
					if not include_hidden and x.startswith("."):
						continue
					lst.add(x[0 : -len(".sccprofile")])
		except OSError:
			pass
	for x in sorted(lst):
		print(x)
	return 0


def cmd_set_profile(argv0: str, argv: list[str]) -> int:
	"""Set controller profile.

	Usage: scc set-profile [controller_id] "profile name"
	"""
	from scc.tools import find_profile

	if len(argv) < 1:
		show_help(command="set_profile", out=sys.stderr)
		return 1
	s = connect_to_daemon()
	if s is None:
		return -1
	if len(argv) >= 2:
		profile = find_profile(argv[1])
		if profile is None:
			print("Unknown profile:", argv[1], file=sys.stderr)
			return 1
		print("Controller: %s" % (argv[0],), file=s)
		if not check_error(s):
			return 1
		print("Profile: %s" % (profile,), file=s)
		if not check_error(s):
			return 1
	else:
		profile = find_profile(argv[0])
		if profile is None:
			print("Unknown profile:", argv[0], file=sys.stderr)
			return 1
		print("Profile: %s" % (profile,), file=s)
		if not check_error(s):
			return 1
	return 0


def cmd_info(argv0: str, argv: list[str]) -> int:
	"""Display basic information about running driver."""
	s = connect_to_daemon()
	if s is None:
		return -1
	# Daemon already sends situable info, so this is mostly reading
	# until "Ready." message is received.
	global_profile = None
	any_controller = False
	while True:
		line = s.readline()
		if len(line) == 0:
			break
		line = line.strip("\r\n\t ")
		if line == "Ready.":
			break
		if line.startswith("Current profile:"):
			global_profile = line
			continue
		if line.startswith("Controller:"):
			continue
		if line.startswith("Controller profile:"):
			any_controller = True
		elif line.startswith("Error:"):
			print(line)
			break
		if ":" in line:
			print(line)
	if not any_controller and global_profile:
		print(global_profile)
	return 0


def cmd_dependency_check(argv0: str, argv: list[str]) -> int:
	"""Check if all required libraries are installed on this system."""
	try:
		import gi

		gi.require_version("Gtk", "3.0")
		gi.require_version("GdkX11", "3.0")
		gi.require_version("Rsvg", "2.0")
	except ValueError as e1:
		print(e1, file=sys.stderr)
		if "Rsvg" in str(e1):
			print("Please, install 'gir1.2-rsvg-2.0' package to use this application", file=sys.stderr)
		else:
			print("Please, install 'PyGObject' package to use this application", file=sys.stderr)
	except ImportError as e2:
		print(e2, file=sys.stderr)
		if "gi" in str(e2):
			print("Please, install 'PyGObject' package to use this application", file=sys.stderr)
		return 1
	try:
		import evdev
	except Exception as e:
		print(e, file=sys.stderr)
		print("Please, install python-evdev package to enable non-steam controller support", file=sys.stderr)
	try:
		import scc.lib.xwrappers as X

		X.Atom
	except Exception as e:
		print(e, file=sys.stderr)
		print("Failed to load X11 helpers, please, check your X installation", file=sys.stderr)
		return 1
	return 0


def cmd_lock_inputs(argv0: str, argv: list[str], lock: str = "Lock: ") -> int:
	"""Lock and print pressed buttons, pads and sticks.

	Locks controller inputs and prints buttons, pads and stick as they are
	pressed or moved on controller.

	Usage: scc lock-inputs [button1] [stick1] [button2] ... [buttonN]

	Available button, sticks and pads:
		A X B Y START C BACK RGRIP LGRIP   LB RB LT RT STICK LPAD RPAD

	Return codes:
		-1  - failed to connect to daemon
		-2  - failed to lock inputs
		-3  - connection terminated
		-4  - daemon reported error
	"""
	s = connect_to_daemon()
	if s is None:
		return -1
	try:
		while True:
			line = s.readline()
			if line == "":
				return -3
			if line.startswith("Ready."):
				print(lock + " ".join([x.upper() for x in argv]), file=s)
				s.flush()
			elif line.startswith("Error:"):
				print(line.strip(), file=sys.stderr)
				return -4
			elif line.startswith("Fail:"):
				print(line.strip(), file=sys.stderr)
				return -2
			elif line.startswith("Event:"):
				data = line.strip().split(" ")
				try:
					print(" ".join(data[2:]), file=sys.stdout)
					sys.stdout.flush()
				except OSError:
					# Output closed, bail out
					return 0
	finally:
		s.close()


def cmd_print_inputs(argv0: str, argv: list[str], lock: str = "Lock: ") -> int:
	"""Print pressed buttons, pads and sticks.

	Prints controller inputs and prints buttons, pads and stick as they are
	pressed or moved on controller, without locking them exclusivelly.

	Usage: scc lock-inputs [button1] [stick1] [button2] ... [buttonN]

	Available button, sticks and pads:
		A X B Y START C BACK RGRIP LGRIP   LB RB LT RT STICK LPAD RPAD

	Return codes:
		-1  - failed to connect to daemon
		-2  - failed to lock inputs
		-3  - connection terminated
		-4  - daemon reported error
	"""
	return cmd_lock_inputs(argv0, argv, lock="Observe: ")


def connect_to_daemon() -> TextIOWrapper | None:
	"""Return socket connected to daemon or None if connection failed.

	Outputs error message in later case.
	"""
	import socket

	from scc.paths import get_daemon_socket

	try:
		s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		s.connect(get_daemon_socket())
	except Exception as e:
		print("Connection to scc-daemon failed: %s" % (e,), file=sys.stderr)
		return None
	return s.makefile(mode="rw")


def check_error(s) -> bool:
	"""Reads line(s) from socket until "OK." or "Fail:" is read.
	Then return True if message is "OK." or prints message to stderr
	and return False if message is "Fail:"
	"""
	s.flush()
	while True:
		line = s.readline()
		if len(line) == 0:
			print("Connection closed", file=sys.stderr)
			return False
		line = line.strip("\n\r\t ")
		if line == "OK.":
			return True
		if line.startswith("Fail:"):
			if "\\n" in line:
				line = line.replace("\\n", "\n")
			print(line, file=sys.stderr)
			return False


def sigint(*a) -> None:
	print("\n*break*")
	sys.exit(0)


def import_osd() -> None:
	import gi

	gi.require_version("Gtk", "3.0")
	gi.require_version("Rsvg", "2.0")
	gi.require_version("GdkX11", "3.0")


def run_osd_tool(tool, argv0: str, argv: list[str]) -> None:
	import signal

	signal.signal(signal.SIGINT, sigint)

	from scc.tools import init_logging

	init_logging()

	sys.argv[0] = "scc osd-keyboard"
	if not tool.parse_argumets([argv0] + argv):
		sys.exit(1)
	tool.run()
	sys.exit(tool.get_exit_code())


def show_help(command=None, out=sys.stdout) -> int:
	names = [x[4:] for x in globals() if x.startswith("cmd_")]
	max_len = max([len(x) for x in names])
	if command in names:
		if "help_" + command in globals():
			return globals()["help_" + command]()
		hlp = (globals()["cmd_" + command].__doc__ or "").strip("\t \r\n")
		if hlp:
			lines = hlp.split("\n")
			if len(lines) > 0:
				for line in lines:
					line = line.replace("Usage: scc", "Usage: %s" % (sys.argv[0],))
					line = line.removeprefix("\t")
					print(line, file=out)
				return 0
	print("Usage: %s <command> [ arguments ]" % (sys.argv[0],), file=out)
	print(file=out)
	print("List of commands:", file=out)
	for name in sorted(names):
		hlp = (globals()["cmd_" + name].__doc__ or "").strip("\t \r\n").split("\n")[0]
		print((" - %%-%ss %%s" % (max_len,)) % (name.replace("_", "-"), hlp), file=out)
	return 0


def main() -> None:
	init_logging()
	if len(sys.argv) < 2:
		sys.exit(show_help())
	if "-h" in sys.argv or "--help" in sys.argv:
		while "-h" in sys.argv:
			sys.argv.remove("-h")
		while "--help" in sys.argv:
			sys.argv.remove("--help")
		sys.exit(show_help(sys.argv[1].replace("-", "_") if len(sys.argv) > 1 else None))
	if "-v" in sys.argv:
		while "-v" in sys.argv:
			sys.argv.remove("-v")
		set_logging_level(True, True)
	else:
		set_logging_level(False, False)
	try:
		command = globals()["cmd_" + sys.argv[1].replace("-", "_")]
	except:
		print("Unknown command: %s" % (sys.argv[1],), file=sys.stderr)
		sys.exit(show_help(out=sys.stderr))

	try:
		sys.exit(command(sys.argv[0], sys.argv[2:]))
	except KeyboardInterrupt:
		sys.exit(0)
	except InvalidArguments:
		print("Invalid arguments", file=sys.stderr)
		print(file=sys.stderr)
		show_help(sys.argv[1], out=sys.stderr)
		sys.exit(1)



#!/usr/bin/env python3
"""SC-Controller - Action Editor - Axis Component

Handles specific XYActions
"""

import logging

from scc.actions import Action, AxisAction, ButtonAction, MouseAction, MultiAction
from scc.gui.ae import AEComponent, describe_action
from scc.gui.area_to_action import action_to_area
from scc.gui.chooser import Chooser
from scc.tools import _

log = logging.getLogger("AE.Axis")

__all__ = ["AxisComponent"]


class AxisComponent(AEComponent, Chooser):
	GLADE = "ae/axis.glade"
	NAME = "axis"
	IMAGES = {"axis": "axistrigger.svg"}
	CTXS = 0

	def __init__(self, app, editor):
		AEComponent.__init__(self, app, editor)
		Chooser.__init__(self, app)
		self.full = None

	def load(self):
		if not self.loaded:
			AEComponent.load(self)
			self.setup_image(grid_columns=2)

	def area_action_selected(self, area, action):
		if area:
			self.set_active_area(area)
		if self.full:
			action = MultiAction(ButtonAction(None, self.full), action)
		self.editor.set_action(action)

	def set_action(self, mode, action):
		if self.handles(mode, action):
			if isinstance(action, MultiAction) and len(action.actions) == 2:
				# axis + button on fully pressed trigger
				self.full = action.actions[0].button2
				self.builder.get_object("lblFullPressed").set_label(
					describe_action(Action.AC_BUTTON, ButtonAction, self.full),
				)
				action = action.actions[1]
			area = action_to_area(action)
			if area is not None:
				self.set_active_area(area)
				return
		self.set_active_area(None)

	def get_button_title(self):
		return _("Trigger or Axis")

	def handles(self, mode, action):
		if isinstance(action, MultiAction) and len(action.actions) == 2:
			# Handles combination of axis + button on fully pressed trigger
			if not isinstance(action.actions[0], ButtonAction):
				return False
			action = action.actions[1]
		return isinstance(action, (AxisAction, MouseAction))



#!/usr/bin/env python3
"""SC-Controller - Icon Chooser
"""

import logging
import os
import re
import traceback

from gi.repository import Gdk, GdkPixbuf, Gio, GObject, Gtk

from scc.gui.dwsnc import headerbar
from scc.gui.editor import Editor
from scc.gui.userdata_manager import UserDataManager
from scc.paths import get_menuicons_path
from scc.tools import _, find_icon

log = logging.getLogger("IconChooser")
RE_URL = re.compile(r"(.*)(https?://[^ ]+)(.*)")
DEFAULT_ICON_CATEGORIES = ("items", "media", "weapons", "system")


class IconChooser(Editor, UserDataManager):
	GLADE = "icon_chooser.glade"

	def __init__(self, app, callback):
		UserDataManager.__init__(self)
		self.app = app
		self.callback = callback
		self.setup_widgets()

	def setup_widgets(self):
		Editor.setup_widgets(self)
		clIcon = self.builder.get_object("clIcon")
		crIconName = self.builder.get_object("crIconName")
		btUserFolder = self.builder.get_object("btUserFolder")
		cr = CellRendererMenuIcon(32)
		clIcon.clear()
		clIcon.pack_start(cr, False)
		clIcon.pack_start(crIconName, True)
		clIcon.set_attributes(cr, icon=1, has_colors=2)
		clIcon.set_attributes(crIconName, text=0)
		btUserFolder.set_label("Add icons...")
		btUserFolder.set_uri("file://%s" % (get_menuicons_path(),))

		headerbar(self.builder.get_object("header"))
		self.load_menu_icons()

	def on_btUserFolder_activate_link(self, *a):
		for c in DEFAULT_ICON_CATEGORIES:
			try:
				os.makedirs(os.path.join(get_menuicons_path(), c))
			except Exception:
				# Dir. exists
				pass

	def on_btOk_clicked(self, *a):
		icon = self.get_selected()
		self.window.destroy()
		if icon:
			self.callback(icon)

	def get_selected(self):
		"""Returns 'category/name' of currently selected icon.
		Returns None if nothing is selected.
		"""
		tsCategories = self.builder.get_object("tsCategories")
		tsIcons = self.builder.get_object("tsIcons")
		try:
			model, iter = tsCategories.get_selected()
			category = model.get_value(iter, 0)
			model, iter = tsIcons.get_selected()
			icon_name = model.get_value(iter, 0)
			return "%s/%s" % (category, icon_name)
		except TypeError:
			# This part may throw TypeError if either list has nothing selected.
			return None

	def on_entName_changed(self, *a):
		pass

	def on_tvItems_cursor_changed(self, view):
		entName = self.builder.get_object("entName")
		lblLicense = self.builder.get_object("lblLicense")
		rvLicense = self.builder.get_object("rvLicense")
		icon = self.get_selected()
		if icon:
			entName.set_text(icon)
			full_path, trash = find_icon(icon)
			if full_path:
				path, name = os.path.split(full_path)
				license = IconChooser.find_license(path, name)
				if license and "(CC 0)" in license:
					# My own icons
					license = license.replace("(CC 0)", "").strip(" ,")
			else:
				license = None
			if license:
				m = RE_URL.match(license)
				if m:
					license = "%s<a href='%s'>%s</a>%s" % (m.group(1), m.group(2), m.group(2), m.group(3))
				lblLicense.set_markup(_("Free-use icon created by %s" % (license,)))
			rvLicense.set_reveal_child(bool(license))

	def on_tvCategories_cursor_changed(self, view):
		model, iter = view.get_selection().get_selected()
		category = model.get_value(iter, 0)
		self.load_menu_icons(category=category)

	@staticmethod
	def color_icon_exists(model, search_name):
		for name, pb, has_colors in model:
			if has_colors and search_name == name:
				return True
		return False

	def on_menuicons_loaded(self, icons):
		tvIcons = self.builder.get_object("tvIcons")
		tvCategories = self.builder.get_object("tvCategories")
		model = tvIcons.get_model()
		model.clear()
		for f in icons:
			name = f.get_basename()
			if (
				f.query_info(Gio.FILE_ATTRIBUTE_STANDARD_TYPE, Gio.FileQueryInfoFlags.NONE, None).get_file_type()
				== Gio.FileType.DIRECTORY
			):
				# ^^ woo, Gio is fun...
				tvCategories.get_model().append((name, name.title()))
			else:
				has_colors = True
				if name.startswith("."):
					# Ignore hidden files
					continue
				name = name.split(".")
				if name[-1] not in ("svg", "png"):
					# Ignore non-supported files
					continue
				name = name[0:-1]
				if name[-1] == "bw":
					has_colors = False
					name = name[0:-1]
				name = ".".join(name)

				if IconChooser.color_icon_exists(model, name):
					continue

				pb = None
				try:
					pb = GdkPixbuf.Pixbuf.new_from_file(f.get_path())
				except Exception as e:
					log.error(e)
					log.error(traceback.format_exc())
					continue

				model.append((name, pb, has_colors))

		trash, selected = tvCategories.get_selection().get_selected()
		if not selected:
			try:
				# Try to select 1st category, but ignore if that fails
				tvCategories.get_selection().select_path((0,))
				self.on_tvCategories_cursor_changed(tvCategories)
			except Exception:
				pass

	@staticmethod
	def find_license(path, name):
		"""Parses LICENSE file, if any, and returns license for give file"""
		licensefile = os.path.join(path, "LICENSES")
		if not os.path.exists(licensefile):
			return None
		with open(licensefile) as file:
			for line in file:
				if line.startswith(name):
					if "-" in line:
						return line.split("-")[-1].strip("\t\r\n ")
		return None


class CellRendererMenuIcon(Gtk.CellRenderer):
	icon = GObject.property(type=GdkPixbuf.Pixbuf)
	has_colors = GObject.property(type=bool, default=False)

	def __init__(self, size):
		Gtk.CellRenderer.__init__(self)
		self.size = size

	def do_get_size(self, *a):
		return 0, 0, self.size, self.size

	def do_render(self, cr, treeview, background_area, cell_area, flags):
		context = Gtk.Widget.get_style_context(treeview)
		Gtk.render_background(
			context, cr, cell_area.x, cell_area.y, cell_area.x + cell_area.width, cell_area.y + cell_area.height,
		)

		scaled = self.icon.scale_simple(self.size, self.size, GdkPixbuf.InterpType.BILINEAR)
		surf = Gdk.cairo_surface_create_from_pixbuf(scaled, 1)
		if self.has_colors:
			cr.set_source_surface(surf, cell_area.x, cell_area.y)
			cr.rectangle(cell_area.x, cell_area.y, self.size, self.size)
		else:
			color_flags = Gtk.StateFlags.NORMAL
			if (flags & Gtk.CellRendererState.SELECTED) != 0:
				color_flags = Gtk.StateFlags.SELECTED
			Gdk.cairo_set_source_rgba(cr, context.get_color(color_flags))
			cr.mask_surface(surf, cell_area.x, cell_area.y)
		cr.fill()
